
/*  ************************************************************************  *
 *				   transfer.c				      *
 *  ************************************************************************  */

/*  This file contains material specific to the int 15h method of moving
    data to and from extended memory.  It accompanies the character device
    driver emulation of a sequential file in extended memory.  */

#include    <driver.h>
#include    <standard.h>

/*  ========================================================================  */

/*  The block's linear address and size must be recorded during device
    driver initialisation (the responsibility of another module).  */

DWORD dwExtMemBase = 0, dwExtMemSize = 0;

/*  Variables describing the file's I/O state are held externally.  */

extern DWORD dwFilePtr, dwFileEnd;		// in seqfile.c

/*  The BIOS provides an int 15h service for moving data to or from extended
    memory.  It requires a Global Descriptor Table (GDT) with space for 6
    segment descriptors, two of which are manipulated by the caller to
    specify source and destination.  */

#pragma pack (1)

static struct SEGMENT_DESCRIPTOR {
    WORD limit;
    WORD base;
    BYTE base_middle_8;
    BYTE access_rights;
    BYTE extra_byte;
    BYTE base_high_8;
} gdt [6] = {0};

/*  ========================================================================  */

/*  The function ToLinear converts a far pointer from segment:offset form to
    an equivalent linear address needed for a segment descriptor.  */

static DWORD ToLinear (void _far *lp)
{
    /*	The calculation is essentially

	    ((DWORD) FP_SEG (lp) << 4) + FP_OFF (lp)

	but must be modified, since the compiler calls a non-resident
	library function to perform bit shifts in 32-bit objects.  This is
	not permissable in a resident module, but note that even without
	this restriction, the following algorithm generates more efficient
	code.  */

    #define FP_SEG(lp)	((WORD) (_segment) (lp))
    #define FP_OFF(lp)	((WORD) (lp))

    return (MK_DWORD (FP_SEG (lp) >> 12, (FP_SEG (lp) << 4) + FP_OFF (lp)));
}

/*  The details of the int 15h facility for transferring data to and from
    extended memory are handled in the following function.  This fills in
    the two descriptors allowed for the source and destination (using linear
    addresses passed as arguments to the function), then issues the int 15h
    instruction to move the data.  No other function need have access to the
    descriptor table.  */

BOOL MoveBlock (BYTE command, BYTE _far *lpBuffer, DWORD xmptr, WORD count)
{
    DWORD source, dest;

    if (command == D_READ) {
	source = dwExtMemBase + xmptr;
	dest = ToLinear (lpBuffer);
    }
    else {
	source = ToLinear (lpBuffer);
	dest = dwExtMemBase + xmptr;
    }

    gdt [2].limit = 0xFFFF;
    gdt [2].base = LOWWORD (source);
    gdt [2].base_middle_8 = LOWBYTE (HIGHWORD (source));
    gdt [2].access_rights = 0x93;
    gdt [2].extra_byte = 0;
    gdt [2].base_high_8 = HIGHBYTE (HIGHWORD (source));

    gdt [3].limit = 0xFFFF;
    gdt [3].base = LOWWORD (dest);
    gdt [3].base_middle_8 = LOWBYTE (HIGHWORD (dest));
    gdt [3].access_rights = 0x93;
    gdt [3].extra_byte = 0;
    gdt [3].base_high_8 = HIGHBYTE (HIGHWORD (dest));

    _asm {
	    push    ds
	    pop     es
	    mov     si,OFFSET gdt	; compiler preserves si
	    mov     cx,count
	    shr     cx,1
	    mov     ah,87h
	    clc
	    int     15h
	    jc	    bad
	    mov     ax,0001h		; TRUE
	    jmp     done
	bad:
	    xor     ax,ax		; FALSE
	done:
    }
}

/*  The function AdjustRequest is called to determine how much of a read or
    write request can actually be met.	Its return value is necessarily less
    than or equal to the DWORD argument.  */

DWORD AdjustRequest (BYTE command, DWORD numwanted)
{
    DWORD avail;

    /*	The block of extended memory obtained by trapping int 15h cannot be
	resized easily, so little adjustment is possible:  simply truncate
	requests to fit the available space.  */

    avail = (command == D_READ ? dwFileEnd : dwExtMemSize) - dwFilePtr;
    if (numwanted > avail) return (avail);
    return (numwanted);
}

/*  ========================================================================  */

/*  No special steps need be taken on opening or closing the device.  */

WORD Open (void)
{
    return (D_DONE);
}

WORD Close (void)
{
    return (D_DONE);
}

/*  ************************************************************************  */

