
/*  ************************************************************************  *
 *				   seqfile.c				      *
 *  ************************************************************************  */

/*  This file contains routines for emulating sequential file I/O with a
    character device driver using extended memory for storage.	*/

#include    <driver.h>
#include    <standard.h>

/*  The "file" is described only by a pointer to the position for the next
    read or write operation and by an end-of-file marker.  */

DWORD dwFilePtr = 0, dwFileEnd = 0;

/*  Services dependent on the nature of extended memory access are obtained
    externally through the following functions.  */

extern DWORD AdjustRequest (BYTE, DWORD);
extern WORD Close (void);
extern BOOL MoveBlock (BYTE, BYTE _far *, DWORD, WORD);
extern WORD Open (void);

/*  ========================================================================  */

/*  There is much similarity between reading and writing, so the device
    driver I/O functions call the following routines, passing a command
    number as the first argument to support the little differentiation
    required.  Both functions return a device driver status word.  */

static WORD ReadWrite (BYTE, BYTE _far *, WORD _far *);
static WORD ReadWriteSingle (BYTE, BYTE _far *);

/*  The first function - the more general of the two - is able to move an
    arbitrary amount of data.  In addition to the device driver status, it
    must also indicate the number of bytes actually transferred, which is
    why the corresponding argument is an address.  */

static WORD ReadWrite (BYTE command, BYTE _far *lpBuffer, WORD _far *lpCount)
{
    register WORD numcopied = 0, status = 0;
    WORD numwanted;

    /*	It might not be possible to accommodate the whole request.  This is
	determined by an external function since the details may depend on
	the method used to obtain the extended memory block.

	A choice exists over the treatment of requests that can be filled
	only partially.  DOS does not copy data to a disk file if space is
	not available to complete the operation, but it might be preferred
	here to truncate requests, leaving it to the caller to decide on
	further action - this is in fact the behaviour adopted here, though
	it does mean that in some cases, an incomplete transfer may be
	reported as successful at user level.  */

    numwanted = (WORD) AdjustRequest (command, *lpCount);

    if (numwanted == 0) status = D_DONE;
    else {

	/*  Both the int 15h service and the more general XMS move data in
	    words, so special provision must be made for odd bytes.  In the
	    meantime, move as much data as possible with one call.  If
	    successful, advance the various pointers and counters, but do
	    not set the DONE bit if a spare byte remains to be transferred.

	    The data transfer is conducted externally:	note that the
	    natural objects for describing the move are the far pointer to
	    the caller's buffer in memory addressable from real mode and an
	    offset into the simulated file.  */

	if (numwanted > 1) {
	    if (MoveBlock (command, lpBuffer, dwFilePtr, numwanted & ~1)) {
		lpBuffer += numwanted & ~1;
		dwFilePtr += numwanted & ~1;
		numcopied += numwanted & ~1;
		if (numcopied == numwanted) status = D_DONE;
	    }
	    else status = D_ERROR | D_DONE | D_GENERAL_FAILURE;
	}

        /*  If the request has not yet been completed, then either a single
            byte must be transferred or a spare byte remains from a larger
            operation.  Either way, the operation is conducted by a separate
            function to lend support for the Non-destructive Read service.  */

	if (NOT (status & D_DONE)) {
	    status = ReadWriteSingle (command, lpBuffer);
	    if (NOT (status & D_ERROR)) {
		dwFilePtr ++;
		numcopied ++;
	    }
	}
    }

    /*	If data has been written past the end-of-file marker, then update
	the marker.  */

    if (dwFilePtr > dwFileEnd) dwFileEnd = dwFilePtr;

    /*  Return the caller's information - the number of bytes successfully
        copied is returned indirectly.  */

    *lpCount = numcopied;
    return (status);
}

/*  The following function moves a single byte and has the same prototype
    as the ReadWrite function except that the argument related to the size
    of the transfer is not required.  Unlike the function above, this one
    does not adjust the "file pointer" after a successful transfer.  It may
    be assumed that 1 byte has indeed been copied if the function returns
    with the ERROR bit clear.  */

static WORD ReadWriteSingle (BYTE command, BYTE _far *lpBuffer)
{
    struct {
	BYTE lowbyte;
	BYTE highbyte;
    } temp;
    DWORD xmtemp;
    register WORD first;

    /*	To transfer a single byte, two bytes must be read into a holding
	buffer so that the one wanted may be either fished out or updated.
	A small complication is introduced by boundary effects:  the first
	byte in the block can only be paired with the one immediately above
	it, the last byte only with the one below it.

	Testing is easier if the first byte is regarded as exceptional, so
	that in all other cases, the byte of interest is paired with the
	one immediately before it in the extended memory block.  */

    xmtemp = dwFilePtr;
    if (dwFilePtr == 0) first = TRUE;
    else {
	first = FALSE;
	xmtemp --;
    }

    /*	Read a pair of bytes from extended memory into the temporary area.  */

    if (MoveBlock (D_READ, (BYTE _far *) &temp, xmtemp, 2) == 0)
	return (D_ERROR | D_DONE | D_GENERAL_FAILURE);

    /*	For a device driver read operation, simply copy the desired byte to
	the caller's buffer.  For a write operation, update the byte in the
	temporary buffer, then move the pair back to extended memory.  */

    if (command == D_READ) *lpBuffer = first ? temp.lowbyte : temp.highbyte;
    else {
	*(first ? &temp.lowbyte : &temp.highbyte) = *lpBuffer;
	if (MoveBlock (D_WRITE, (BYTE _far *) &temp, xmtemp, 2) == 0)
	    return (D_ERROR | D_DONE | D_GENERAL_FAILURE);
    }
    return (D_DONE);
}

/*  ========================================================================  */

/*  Device driver functions - in the form required for DRIVER.ASM  */

/*  Read and write functions transfer the data between the caller's buffer
    and a region of extended memory beginning some distance into the
    extended memory block.  They have much in common, so in both cases,
    extract relevant fields from the request packet and pass them to a
    common function.  */

WORD DrvrProcRead (struct REQUEST_HEADER _far *lpReqHdr)
{
    return (ReadWrite (D_READ, lpReqHdr -> lpBuffer, &(lpReqHdr -> wCount)));
}

WORD DrvrProcWrite (struct REQUEST_HEADER _far *lpReqHdr)
{
    return (ReadWrite (D_WRITE, lpReqHdr -> lpBuffer, &(lpReqHdr -> wCount)));
}

/*  The Write With Verify command is treated no differently from the normal
    Write command.  */

WORD DrvrProcWriteWithVerify (struct REQUEST_HEADER _far *lpReqHdr)
{
    return (DrvrProcWrite (lpReqHdr));
}

/*  For the Non-destructive Read, first check that a byte is available, then
    call the separate function used for picking up spares.  The Input Status
    command is the same but does not return the waiting character.  Note
    however, that DOS never issues this command:  int 21h function 4406h -
    described as Input Status - is actually a Non-destructive Read at device
    driver level.  */

WORD DrvrProcNondestructiveRead (struct REQUEST_HEADER _far *lpReqHdr)
{
    if (AdjustRequest (D_READ, 1) == 1)
	return (ReadWriteSingle (D_READ, &(lpReqHdr -> cCharWaiting)));
    else return (D_ERROR | D_DONE | D_GENERAL_FAILURE);
}

WORD DrvrProcInputStatus (struct REQUEST_HEADER _far *lpReqHdr)
{
    if (AdjustRequest (D_READ, 1) == 1) return (D_DONE);
    else return (D_ERROR | D_DONE | D_GENERAL_FAILURE);
}

WORD DrvrProcOutputStatus (struct REQUEST_HEADER _far *lpReqHdr)
{
    if (AdjustRequest (D_WRITE, 1) == 1) return (D_DONE);
    else return (D_ERROR | D_DONE | D_GENERAL_FAILURE);
}

/*  ------------------------------------------------------------------------  */

/*  The Open Device command may be used to ensure that the file pointer is
    reset to zero whenever the device is opened.  Among other things, this
    supports storage and retrieval across successive openings.

    The function given below for closing the device truncates the "file" at
    the position of the last read or write.

    In both cases, the function is also passed externally to support
    operations that depend on the nature of the extended memory allocation.

    No attempt is made here to address the question of multiple ownership.  */

WORD DrvrProcOpenDevice (struct REQUEST_HEADER _far *lpReqHdr)
{
    dwFilePtr = 0;
    return (Open ());
}

WORD DrvrProcCloseDevice (struct REQUEST_HEADER _far *lpReqHdr)
{
    dwFileEnd = dwFilePtr;
    return (Close ());
}

/*  ------------------------------------------------------------------------  */

/*  The IOCTL interface - intended for passing control data to and from the
    device - may be used to get and set the "file pointer".  In the simple
    implementation which follows, this pointer - a 32-bit offset into the
    extended memory block - is passed as a DWORD in the IOCTL buffer.  */

WORD DrvrProcIOCTLRead (struct REQUEST_HEADER _far *lpReqHdr)
{
    /*	If the transfer count is the right size, copy the "file pointer"
	into the buffer.  */

    if (lpReqHdr -> wCount == sizeof (dwFilePtr))
	*((DWORD _far *) lpReqHdr -> lpBuffer) = dwFilePtr;
    else lpReqHdr -> wCount = 0;
    return (D_DONE);
}

WORD DrvrProcIOCTLWrite (struct REQUEST_HEADER _far *lpReqHdr)
{
    DWORD temp;

    /*	If the caller indicates that the buffer contains exactly four bytes
	of control data, interpret these bytes as a new value for the file
	pointer.  */

    if (lpReqHdr -> wCount == sizeof (dwFilePtr)) {

	/*  Accept a lower value for the file pointer but allow an advance
	    only if it would be possible to write that far into the file.  */

	temp = *((DWORD _far *) lpReqHdr -> lpBuffer);
	if (temp <= dwFilePtr) dwFilePtr = temp;
	else {
	    temp -= dwFilePtr;
	    if (AdjustRequest (D_WRITE, temp) == temp) dwFilePtr += temp;
	    else lpReqHdr -> wCount = 0;
	}
    }
    else lpReqHdr -> wCount = 0;
    return (D_DONE);
}

/*  ************************************************************************  */

