
/*  ************************************************************************  *
 *				    diskid.c				      *
 *  ************************************************************************  */

#include    <dos.h>
#include    <stdio.h>
#include    <string.h>

#include    <doserror.h>
#include    <drivparm.h>
#include    <standard.h>

/*  This file provides the DISKDUP utility with functions that deal with
    disk identification.  */

BOOL GetDiskCharacteristics (WORD, struct DRIVPARM *);
BOOL GetDiskMediaID (WORD, struct MEDIAID *, BOOL *);
BOOL HasDiskChanged (WORD drive, struct MEDIAID *idptr);
BOOL SetDiskMediaID (WORD, struct MEDIAID *, BOOL *);

/*  Use int 21h function 440Dh to access storage media below the level of
    the file system.  This allows DISKDUP to work with local drives that are
    not supported by the int 13h BIOS services.  The following C functions
    each correspond to an int 21h call and follow the Microsoft C scheme of
    returning the DOS error code on failure or zero on success.  Like most
    int 21h functions, these expect 1-based drive numbers.  */

extern WORD _dos_getdrivparm (WORD, struct DRIVPARM *);     /*	gioctl.c  */
extern WORD _dos_getmediaid (WORD, struct MEDIAID *);	    /*	gioctl.c  */
extern WORD _dos_setmediaid (WORD, struct MEDIAID *);	    /*	gioctl.c  */

/*  A function to describe DOS error values in a standard form	*/

extern CHAR *GetErrormsg (WORD);			    /*	errormsg.c  */

/*  Some self-explanatory functions declared in MAIN.C	*/

extern void PromptForKey (void);
extern void PutError (CHAR *, ...);
extern BOOL RetryError (CHAR *, ...);

/*  ========================================================================  */

/*  The following function fills a DRIVPARM structure about the current
    contents of a designated drive.  The function prompts for a disk, if the
    drive is empty.  */

BOOL GetDiskCharacteristics (WORD drive, struct DRIVPARM *parmptr)
{
    register WORD driveletter, errorcode;
    struct DOSERROR errorinfo;

    driveletter = drive + 'A';

    /*	Note that any attempt at obtaining disk characteristics may have
	to be repeated if the drive is empty, hence the for loop.  */

    for (;;) {
	parmptr -> cSpecial = DRIVPARM_CURRENT;
	if (_dos_getdrivparm (drive + 1, parmptr) == 0) return (TRUE);

	/*  The normal DOS error code returned by the Generic IOCTL function
	    is not sufficiently informative - note in particular that the
	    value 0x0005 covers all device errors, whatever their cause - so
	    obtain extended error information.	*/

	errorcode = dosexterr (&errorinfo);

	/*  If disk characteristics cannot be obtained because the drive
	    is undefined or remote or because the responsible device driver
	    does not offer a Generic IOCTL interface, deal with the error
	    here.  */

	if (errorcode == E_INVALID_DRIVE) {
	    PutError ("drive %c: is invalid", driveletter);
	    return (FALSE);
	}
	if (errorcode == E_INVALID_FUNCTION OR errorcode == E_BAD_COMMAND) {
	    PutError ("drive %c: is unsuitable", driveletter);
	    return (FALSE);
	}

	/*  If the problem is simply that the drive contains no disk, prompt
	    for one and try again.  Strictly speaking, the user should not
	    be asked to insert a disk until the program has first checked
	    that media replacement is possible.  */

	if (errorcode == E_NOT_READY) {
	    printf ("\nInsert disk in drive %c:", driveletter);
	    PromptForKey ();
	    continue;
	}

	/*  For any other problem, describe the error in some detail and
	    offer the chance to retry.	Note that device errors during IOCTL
	    services do not generate int 24h exceptions - the message
	    displayed by the RetryError function is not too dissimilar
	    however to the familiar "Abort, Retry, Ignore, Fail".  */

	if (NOT RetryError ("%s reading drive %c:",
		GetErrormsg (errorcode), driveletter)) return (FALSE);
    }
}

/*  ------------------------------------------------------------------------  */

/*  Two functions are presented for dealing with media identifiers (volume
    label, serial number and file system).  Both functions pass the 0-based
    drive number and two addresses:  the first points to a MEDIAID
    structure whose validity is indicated by the Boolean flag at the second
    address.

    The functions return FALSE if a device driver error prevents
    determination of the existence of a media id for the disk in the
    designated drive or of the ability of the disk format to accept a new
    media id.  */

BOOL GetDiskMediaID (WORD drive, register struct MEDIAID *idptr, BOOL *flagptr)
{
    struct DOSERROR errorinfo;
    WORD errorcode;

    for (;;) {
	idptr -> wInfoLevel = 0;
	if (_dos_getmediaid (drive + 1, idptr) == 0) {
	    *flagptr = TRUE;
	    return (TRUE);
	}
	*flagptr = FALSE;

	errorcode = dosexterr (&errorinfo);
	if (errorcode == E_INVALID_FUNCTION OR errorcode == E_NOT_DOS_DISK)
	    return (TRUE);

	if (NOT RetryError ("%s reading media id drive %c:",
		GetErrormsg (errorcode), drive + 'A')) return (FALSE);
    }
}

BOOL SetDiskMediaID (WORD drive, register struct MEDIAID *idptr, BOOL *flagptr)
{
    struct dosdate_t date;
    struct dostime_t time;
    struct DOSERROR errorinfo;
    WORD errorcode;
    DWORD jumbledtime;

    for (;;) {

	/*  This function adds a 'unique' serial number to the other media
	    identifiers, sparing the caller from the details of concocting
	    something suitable.  Microsoft's choice of structures for the
	    C library routines that give access to the DOS Get Date/Time
	    functions do not lend themselves immediately to creating a
	    serial number compatible with the standard FORMAT program.
	    That utility simply adds the cx values returned by the two
	    functions to get the high word of the new serial number, then
	    does the same with the dx values to get the low word.  The same
	    can be achieved easily enough here with a little jumbling.	*/

	_dos_getdate (&date);
	_dos_gettime (&time);
	jumbledtime = MK_DWORD (MK_WORD (time.hour, time.minute),
			MK_WORD (time.second, time.hsecond));
	idptr -> dwSerialNumber = *((DWORD *) &date) + jumbledtime;
	idptr -> wInfoLevel = 0;
	if (_dos_setmediaid (drive + 1, idptr) == 0) {
	    *flagptr = TRUE;
	    return (TRUE);
	}
	*flagptr = FALSE;

	/*  In general, alert the user to an error, but note that there is
	    not much point to this if the problem lies with a disk format
	    that doesn't support the media identifiers.  */

	errorcode = dosexterr (&errorinfo);
	if (errorcode == E_INVALID_FUNCTION OR errorcode == E_NOT_DOS_DISK)
	    return (TRUE);

	if (NOT RetryError ("%s writing media id drive %c:",
		GetErrormsg (errorcode), drive + 'A')) return (FALSE);
    }
}

/*  ------------------------------------------------------------------------  */

BOOL HasDiskChanged (WORD drive, register struct MEDIAID *idptr)
{
    struct MEDIAID currentid;
    BOOL flag;

    /*	The argument idptr (if not NULL) addresses a structure containing
	the media identifiers expected of the disk in the specified drive.
	Comparison with the identifiers belonging to the disk that actually
	is in the drive may show that the drive contents have changed (in
	which case the function returns TRUE).

	In the first instance however, return FALSE (meaning don't know) if
	some sort of disk error prevents determination even of the existence
	of media identifiers.  */

    if (NOT GetDiskMediaID (drive, &currentid, &flag)) return (FALSE);

    /*	If the current contents of the drive do not offer a media id, give
	benefit of doubt if identifiers were not expected, but otherwise
	regard the disk as changed.  */

    if (NOT flag) return (idptr == NULL ? FALSE : TRUE);

    /*	Given that media identifiers exist now, regard the disk as changed
	if there were no identifiers expected or if the current identifiers
	differ from the expected ones.	*/

    if (idptr == NULL) return (TRUE);
    if (memcmp (&currentid, idptr, sizeof (*idptr))) return (TRUE);

    /*	That the current contents have the expected media id does not prove
	that no disk change has occurred.  */

    return (FALSE);
}

/*  ************************************************************************  */

