
/*  ************************************************************************  *
 *				    diskio.c				      *
 *  ************************************************************************  */

#include    <dos.h>
#include    <stdio.h>
#include    <string.h>

#include    <doserror.h>
#include    <drivparm.h>
#include    <standard.h>

/*  This file provides the DISKDUP utility with functions that arrange disk
    access.  */

BOOL ReadDisk (WORD, struct DRIVPARM *, void *);
BOOL WriteDisk (WORD, struct DRIVPARM *, void *);

static BOOL ReadTrack (WORD, WORD, register struct RWBLOCK *);
static BOOL WriteTrack (WORD, WORD, register struct RWBLOCK *);
static void SetDrivparm (WORD, struct DRIVPARM *);

extern BOOL GetDiskMediaID (WORD, struct MEDIAID *, BOOL *);
extern BOOL HasDiskChanged (WORD drive, struct MEDIAID *idptr);
extern BOOL SetDiskMediaID (WORD, struct MEDIAID *, BOOL *);

extern WORD _dos_getdrivparm (WORD, struct DRIVPARM *);     /*	gioctl.c  */
extern WORD _dos_setdrivparm (WORD, struct DRIVPARM *);     /*	gioctl.c  */
extern WORD _dos_readtrack (WORD, struct RWBLOCK *);	    /*	gioctl.c  */
extern WORD _dos_writetrack (WORD, struct RWBLOCK *);	    /*	gioctl.c  */

extern BOOL bSetDrivparm;				    /*	main.c	*/

extern struct MEDIAID SrcID, DestID;			    /*	main.c	*/
extern BOOL bSrcID, bDestID;				    /*	main.c	*/

static BOOL ChangeDestID (WORD, struct MEDIAID *, BOOL);
static BOOL bDestIDChanged;

/*  Questions about where the material read from the source disk is held
    until the target disk is ready are all dealt with externally.  This
    module need know only a few abstract operations.  */

extern BOOL ArrangeStorage (DWORD);			    /*	storage.c  */
extern WORD ReadStorage (void _far *, WORD);		    /*	storage.c  */
extern WORD WriteStorage (void _far *, WORD);		    /*	storage.c  */
extern BOOL RewindStorage (void);			    /*	storage.c  */

/*  A function to describe DOS error values in a standard form	*/

extern CHAR *GetErrormsg (WORD);			    /*	errormsg.c  */

/*  Some self-explanatory functions declared in MAIN.C	*/

extern void PromptForKey (void);
extern void PutError (CHAR *, ...);
extern BOOL RetryError (CHAR *, ...);

/*  ========================================================================  */

/*  Although the program never alters the characteristics of a disk, it
    still requires the GIOCTL service that sets drive parameters.  It isn't
    stated very clearly in the official literature, but for a number of
    reasons, mostly concerning undocumented track layout fields, the Generic
    IOCTL commands for reading and writing tracks, as implemented for the
    default block device driver in IO.SYS, are prone to peculiar behaviour
    if the device parameters are not set first.  Refer to the accompanying
    text for details.  */

static void SetDrivparm (WORD drive, struct DRIVPARM *parmptr)
{
    register WORD sector;
    register struct SECTORPARM *secptr;

    /*	It is assumed here that the DRIVPARM structure already contains a
	BPB for the media and that the only problem is track layout.  */

    parmptr -> cSpecial = SET_TRACKLAYOUT_ONLY | SET_GOOD_TRACKLAYOUT;

    #ifdef _MINIMAL_

    /*	A minimal solution is to assume that the default track layout table
	is still in place - that no operations are being conducted on disks
	with unusual sector sizes, let alone disks that have variations
	within a track.  In this case, there is an easy way to have the
	Set Device Parameters command avoid loading a track layout table.  */

    parmptr -> wSectorsInTable = 0;

    #else

    /*	For safety however, the track layout table should be refreshed.  */

    parmptr -> wSectorsInTable = parmptr -> bpb.wSectorsPerTrack;

    secptr = parmptr -> tracklayout;
    sector = 1;
    do {
	secptr -> wNumber = sector;
	secptr -> wSize = 512;
	secptr ++;
	sector ++;
    } while (sector < parmptr -> wSectorsInTable);

    #endif

    _dos_setdrivparm (drive + 1, parmptr);
}

/*  ------------------------------------------------------------------------  */

BOOL ReadDisk (WORD drive, struct DRIVPARM *parmptr, void *buffer)
{
    register WORD bytes_per_track, heads;
    WORD cylinders, sectors_per_track;
    DWORD total_bytes, total_sectors;
    struct RWBLOCK rwblock;

    if (bSetDrivparm) SetDrivparm (drive, parmptr);

    /*	Calculate certain useful quantities.  */

    heads = parmptr -> bpb.wHeads;
    sectors_per_track = parmptr -> bpb.wSectorsPerTrack;
    bytes_per_track = sectors_per_track * parmptr -> bpb.wBytesPerSector;
    total_sectors = parmptr -> bpb.wSectors ?
	    (DWORD) parmptr -> bpb.wSectors : parmptr -> bpb.dwHugeSectors;
    cylinders = (WORD) (total_sectors / (sectors_per_track * heads));
    total_bytes = total_sectors * parmptr -> bpb.wBytesPerSector;

    /*	Prepare the necessary storage space, making sure that enough of it
	exists to hold the contents of the entire disk.  */

    if (NOT ArrangeStorage (total_bytes)) {
	PutError ("insufficient temporary storage for entire disk");
	return (FALSE);
    }

    /*	Prepare the RWBLOCK structure for the Generic IOCTL command that
	reads disk tracks, then loop through the disk, track by track.	*/

    rwblock.cSpecial = 0;
    rwblock.wCylinder = 0;
    rwblock.wFirstSector = 0;
    rwblock.wSectors = sectors_per_track;
    rwblock.lpBuffer = (void _far *) buffer;

    do {
	rwblock.wHead = 0, heads = parmptr -> bpb.wHeads;
	do {
	    if (NOT ReadTrack (drive, bytes_per_track, &rwblock))
		return (FALSE);
	} while (++ rwblock.wHead, -- heads);
    } while (++ rwblock.wCylinder, -- cylinders);

    return (TRUE);
}

BOOL WriteDisk (WORD drive, struct DRIVPARM *parmptr, void *buffer)
{
    register WORD bytes_per_track, heads;
    WORD cylinders, sectors_per_track;
    DWORD total_bytes, total_sectors;
    struct RWBLOCK rwblock;

    if (bSetDrivparm) SetDrivparm (drive, parmptr);

    heads = parmptr -> bpb.wHeads;
    sectors_per_track = parmptr -> bpb.wSectorsPerTrack;
    bytes_per_track = sectors_per_track * parmptr -> bpb.wBytesPerSector;
    total_sectors = parmptr -> bpb.wSectors ?
	    (DWORD) parmptr -> bpb.wSectors : parmptr -> bpb.dwHugeSectors;
    cylinders = (WORD) (total_sectors / (sectors_per_track * heads));
    total_bytes = total_sectors * parmptr -> bpb.wBytesPerSector;

    /*	The storage space is accessed sequentially - now that its contents
	must be retrieved, any internal position counters describing that
	storage will have to be rewound to the start.  */

    if (NOT RewindStorage ()) {
	PutError ("error controlling temporary storage");
	return (FALSE);
    }

    bDestIDChanged = FALSE;

    rwblock.cSpecial = 0;
    rwblock.wCylinder = 0;
    rwblock.wFirstSector = 0;
    rwblock.wSectors = sectors_per_track;
    rwblock.lpBuffer = (void _far *) buffer;

    do {
	rwblock.wHead = 0, heads = parmptr -> bpb.wHeads;
	do {
	    if (NOT WriteTrack (drive, bytes_per_track, &rwblock))
		return (FALSE);
	} while (++ rwblock.wHead, -- heads);
    } while (++ rwblock.wCylinder, -- cylinders);

    return (TRUE);
}

/*  ------------------------------------------------------------------------  */

static BOOL ReadTrack (WORD drive, WORD tracksize,
			    register struct RWBLOCK *rwptr)
{
    struct DOSERROR errorinfo;

    /*	Read the track indicated by fields in the RWBLOCK structure passed
	as one of the arguments.  */

    while (_dos_readtrack (drive + 1, rwptr) != 0) {

	/*  If there is an error, alert the user and ask to retry the
	    operation.	*/

	if (NOT RetryError ("%s reading track %d head %d",
		GetErrormsg (dosexterr (&errorinfo)),
		rwptr -> wCylinder, rwptr -> wHead))
	    return (FALSE);

	/*  Before re-attempting the track read, try to establish that the
	    original source disk has not been swapped for another during
	    the pause - there's no explaining what some people do with their
	    computers!	*/

	if (HasDiskChanged (drive, bSrcID ? &SrcID : NULL)) {
	    PutError ("unexpected disk change");
	    return (FALSE);
	}
    }

    /*	Having read the track successfully, append the data to the rest that
	has been read from disk, checking that the full amount is saved
	correctly.  */

    if (WriteStorage (rwptr -> lpBuffer, tracksize) != tracksize) {
	PutError ("storage failure while saving track %d head %d",
	    rwptr -> wCylinder, rwptr -> wHead);
	return (FALSE);
    }
    return (TRUE);
}

static BOOL WriteTrack (WORD drive, WORD tracksize,
			    register struct RWBLOCK *rwptr)
{
    struct DOSERROR errorinfo;
    struct MEDIAID id;
    BOOL idflag;

    /*	This is essentially the reverse of the previous function - but with
	some small exceptions.	*/

    if (ReadStorage (rwptr -> lpBuffer, tracksize) != tracksize) {
	PutError ("storage failure while retrieving track %d head %d",
	    rwptr -> wCylinder, rwptr -> wHead);
	return (FALSE);
    }
    while (_dos_writetrack (drive + 1, rwptr) != 0) {
	if (NOT RetryError ("%s writing track %d head %d",
		GetErrormsg (dosexterr (&errorinfo)),
		rwptr -> wCylinder, rwptr -> wHead))
	    return (FALSE);
	if (HasDiskChanged (drive, bDestID ? &DestID : NULL)) {
	    PutError ("unexpected disk change");
	    return (FALSE);
	}
    }

    /*	If either the source disk or the destination disk possessed media
	identifiers at the start, then at some stage during the copy
	operation, the destination disk's identifiers will change.  Most
	probably, the change will occur when the first track is written,
	but this does not appear to be a formal requirement for media
	identifiers.

	The change has to be caught for two reasons:  first, to support
	detection of a disk change during an option to retry and second, to
	write a distinct serial number to the destination disk at the
	earliest possible opportunity.	*/

    if (NOT bDestIDChanged AND (bSrcID OR bDestID)) {
	if (NOT GetDiskMediaID (drive, &id, &idflag)) return (FALSE);
	bDestIDChanged = ChangeDestID (drive, &id, idflag);
    }

    return (TRUE);
}

static BOOL ChangeDestID (WORD drive, struct MEDIAID *idptr, BOOL idflag)
{
    if (idflag) {
	if (bSrcID) {

	    /*	In this case, there is currently an id and there is
		eventually meant to be (by the time the source disk has been
		copied completely).  If the destination disk had an id all
		along and this differs from the source disk's, then the new
		id has not been written to the destination yet.  Return
		FALSE so that the check will be repeated.  */

	    if (bDestID AND memcmp (idptr, &SrcID, sizeof (SrcID)))
		return (FALSE);

	    /*	Now it is clear that the destination disk has acquired the
		id from the source.  Update the DestID structure so that any
		future disk changes will be caught and write a distinct
		serial number to the destination disk.	*/

	    memcpy (&DestID, idptr, sizeof (*idptr));
	    return (SetDiskMediaID (drive, &DestID, &bDestID));
	}
	else {

	    /*	For this case, there is currently an id on the destination
		disk, but there should not be once the copy operation is
		complete.  Regard this situation as indicating that the
		original media id on the destination disk has not yet been
		overwritten.  */

	    return (FALSE);
	}
    }
    else {
	if (bSrcID) return (FALSE);
	else {
	    bDestID = FALSE;
	    return (TRUE);
	}
    }
}

/*  ************************************************************************  */

