
/*  ************************************************************************  *
 *		    main.c - primary module for diskdup.exe		      *
 *  ************************************************************************  */

#include    <conio.h>
#include    <dos.h>
#include    <malloc.h>
#include    <stdarg.h>
#include    <stdio.h>
#include    <stdlib.h>
#include    <string.h>

#include    <doserror.h>
#include    <drivparm.h>
#include    <standard.h>

static CHAR description [] = "\n\
Duplicates a floppy disk in the same drive, using conventional\n\
and extended memory to avoid tedious disk changes.\n\
\n\
  DISKDUP [drive:]\n\
";

/*  The program uses int 21h function 440Dh services to access storage media
    at a level below the file system.  The details are handled externally -
    mostly in DISK.C - but this module must supply some relevant structures
    and flags to keep track of the relationship between source and
    destination disks.	*/

extern BOOL GetDiskCharacteristics (WORD, struct DRIVPARM *);	/*  disk.c  */
extern BOOL GetDiskMediaID (WORD, struct MEDIAID *, BOOL *);	/*  disk.c  */
extern BOOL ReadDisk (WORD, struct DRIVPARM *, void *); 	/*  disk.c  */
extern BOOL WriteDisk (WORD, struct DRIVPARM *, void *);	/*  disk.c  */

struct MEDIAID SrcID, DestID;
BOOL bSrcID, bDestID;

struct DRIVPARM SrcParm, DestParm;

BOOL bSetDrivparm = TRUE;

void DrainStdin (void);
BOOL GetYesNo (void);
void PromptForKey (void);
void PutError (CHAR *, ...);
void PutNewLine (void);
BOOL RetryError (CHAR *, ...);

/*  ========================================================================  */

int main (int argc, char **argv)
{
    register CHAR *ptr;
    WORD drive, driveletter, tracksize;
    BOOL drivegiven = FALSE;
    void *buffer;

    /*	Parse the command line.  */

    while (++ argv, -- argc) {
	ptr = *argv;
	if (*ptr == '/') {
	    ptr ++;

	    /*	A switch is provided to demonstrate a shortcoming of the
		official documentation for Generic IOCTL services - refer to
		the accompanying text for details.  Any initial substring is
		accepted, irrespective of case.  */

	    if (strnicmp ("SetDrivparmOff", ptr, strlen (ptr)) == 0) {
		bSetDrivparm = FALSE;
		continue;
	    }

	    /*	A description of the program may be obtained by supplying
		the /? switch as usual.  */

	    if (ptr [0] == '?' AND ptr [1] == '\0') {
		printf (description);
		return (0);
	    }

	    /*	No other switches are accepted.  Exit codes, incidentally,
		conform loosely to those of DISKCOPY.  */

	    PutError ("invalid switch - %s", -- ptr);
	    return (4);
	}
	else {

	    /*	The drive containing the disk to be duplicated may be given
		in upper or lower case, but as usual, must be followed
		immediately by a colon and nothing else.  Do not allow the
		user to specify two drives.  Record the drive as a 0-based
		number.  */

	    if (drivegiven) {
		PutError ("too many arguments");
		return (4);
	    }
	    driveletter = ptr [0] & ~0x20;
	    drive = driveletter - 'A';
	    if (drive <= 'Z' - 'A' AND ptr [1] == ':' AND ptr [2] == '\0') {
		drivegiven = TRUE;
		continue;
	    }

	    /*	No other parameters are accepted.  */

	    PutError ("invalid parameter - %s", ptr);
	    return (4);
	}
    }

    /*	If no drive was specified on the command line, call DOS to determine
	the default drive.  Most int 21h functions work with 1-based drive
	numbers.  In this program, drive numbers are kept in 0-based form
	except when calling DOS functions (or direct C counterparts in the
	_dos style of the Microsoft C library).  */

    if (NOT drivegiven) {
	_dos_getdrive (&drive);
	drive --;
	driveletter = drive + 'A';
    }

    /*	Determine the nature of the disk in the nominated drive, possibly
	also prompting for a disk or displaying a fatal error message if the
	disk or drive is unacceptable.	Those choices however are hidden
	from this level of the program's modular heirarchy.  */

    if (NOT GetDiskCharacteristics (drive, &SrcParm)) return (4);

    /*	Reject attempts to duplicate fixed disks (a design decision rather
	than a lower-level error).  */

    if (SrcParm.wDeviceAttributes & DRIVE_FIXED) {
	PutError ("drive %c: contains non-removable media", driveletter);
	return (4);
    }

    /*	While still investigating the source disk's properties, record its
	volume label and serial number.  The GIOCTL interface works beneath
	the DOS file system and the program is therefore vulnerable to
	unanticipated disk changes (as is the standard DISKCOPY utility).
	The media identifiers, particularly the serial number introduced
	with DOS 4, may help the program trap accidents or silly behaviour.  */

    if (NOT GetDiskMediaID (drive, &SrcID, &bSrcID)) return (4);

    /*	Allocate a buffer from the heap to hold successive tracks read from
	disk before they are passed to more substantial storage.  The same
	buffer will be used when the process is reversed to make the
	duplicate disk.  */

    tracksize = SrcParm.bpb.wBytesPerSector * SrcParm.bpb.wSectorsPerTrack;
    buffer = malloc (tracksize);
    if (buffer == NULL) {
	PutError ("insufficient conventional memory");
	return (4);
    }

    /*	Read the disk a track at a time, copying each track into whatever
	temporary storage is available - whether this be conventional memory,
	extended memory or some file on a hard disk, or a combination of
	these (the details are of no concern at this level).  */

    printf ("\nReading disk, please wait ...\n");
    if (NOT ReadDisk (drive, &SrcParm, buffer)) return (3);

    /*	Ask for a second disk, then obtain its characteristics for
	comparison with those of the source.  */

    for (;;) {
	printf ("\n\nInsert target disk in drive %c:", driveletter);
	for (;;) {
	    PromptForKey ();
	    if (NOT GetDiskCharacteristics (drive, &DestParm)) return (4);

	    /*	In this simple implementation, the target disk must already
		be formatted to the same specifications as the source disk.
		Should this not be so, alert the user.	Further in the for
		loop there is a message offering the user the chance to
		replace the disk.  */

	    if (memcmp (&DestParm.bpb, &SrcParm.bpb, sizeof (struct BPB))) {
		printf ("\nTarget disk must have same format as source.");
	    }
	    else {

		/*  If the source and target disks do have the same format,
		    guard against the possibility that they are the same
		    disk - that the user has not answered the prompt for a
		    new disk without actually having changed disks.  */

		if (NOT GetDiskMediaID (drive, &DestID, &bDestID)) return (4);

		if (NOT (bDestID AND bSrcID)) break;
		if (memcmp (&DestID, &SrcID, sizeof (SrcID)) != 0) break;

		printf ("\nTarget disk has same serial number as source.");
		printf ("\nIs this disk intended as the target (y/n)?  ");

		if (GetYesNo ()) {
		    PutNewLine ();
		    break;
		}
		PutNewLine ();
	    }
	    printf ("\nInsert replacement disk in drive %c:", driveletter);
	}

	/*  Retrieve the contents of the source disk from storage and copy
	    them to the new disk in the drive (again, the details are of no
	    concern at this level).  */

	printf ("\nWriting disk, please wait ...\n");
	if (WriteDisk (drive, &DestParm, buffer)) {
	    printf ("\n\nDisk duplicated ok.  Make another copy (y/n)?  ");
	    if (NOT GetYesNo ()) {
		PutNewLine ();
		return (0);
	    }
	}
	else {
	    printf ("\n\nDuplication failure.  Try another disk (y/n)?  ");
	    if (NOT GetYesNo ()) {
		PutNewLine ();
		return (3);
	    }
	}
    }
}

/*  ------------------------------------------------------------------------  */

/*  The remaining functions provide uniform treatment of console I/O for
    prompts and error messages.  */

void DrainStdin (void)
{
    while (kbhit ()) {
	getch ();
    }
}

BOOL GetYesNo (void)
{
    WORD ch;

    for (;;) {
	ch = getch ();

	/*  Even for something as simple as waiting for a y/n response,
	    remember that special function keys, Alt combinations, etc are
	    signalled by two bytes - if the first byte is null, do not leave
	    the second in the console buffer to confuse other routines or
	    programs.  */

	if (ch == 0) {
	    getch ();
	    continue;
	}
	if (ch == 'y'OR ch == 'Y') {
	    putch (ch);
	    return (TRUE);
	}
	if (ch == 'n' OR ch == 'N') {
	    putch (ch);
	    return (FALSE);
	}
    }
}

void PromptForKey (void)
{
    printf ("\nPress any key when ready ...\n");
    DrainStdin ();
    while (getch () == 0) {
    }
}

void PutError (CHAR *str, ...)
{
    va_list va_start (argptr, str);
    printf ("\nDISKDUP fatal error:  ");
    vprintf (str, argptr);
    PutNewLine ();
    va_end (argptr);
}

void PutNewLine (void)
{
    printf ("\n");
}

BOOL RetryError (CHAR *str, ...)
{
    register WORD answer;
    va_list va_start (argptr, str);

    vprintf (str, argptr);
    va_end (argptr);
    printf ("\nRetry (y/n)?  ");
    answer = GetYesNo ();
    PutNewLine ();
    return (answer);
}

/*  ************************************************************************  */

