
/*  ************************************************************************  *
 *				   biosdrv.c				      *
 *  ************************************************************************  */

/*  This file demonstrates two methods to determine the int 13h drive number
    corresponding to a DOS drive (0-based throughout the file).  */

#include    <bios.h>
#include    <dos.h>
#include    <stdlib.h>

#include    <intpack.h>
#include    <iosys.h>
#include    <standard.h>

/*  In this first part of the file, simply select the method.

    The first, which is available only with recent DOS versions, uncovers
    the drive number used at int 13h level only if the DOS drive is managed
    by the default block device driver in IO.SYS.

    The second method installs a temporary int 13h handler so that the drive
    number used for int 13h access can be recorded when a sector is read
    from the DOS drive.  It uses only documented APIs but has all the usual
    disadvantages of invasive testing, made most noticable in this case by
    the need to "hit the disk".  */

static BYTE UseIOSYSDriveTables (WORD);     // first method
static BYTE UseInt13hTrap (WORD);	    // second method

BYTE GetBIOSDrive (WORD dosdrive)
{
    #ifdef _NO_IOSYS_			    // for testing purposes, compile
    return (UseInt13hTrap (dosdrive));	    // with the _NO_IOSYS_ symbol to
    #else				    // force use of int 13h trap

    BYTE biosdrive;

    /*	If the first method doesn't work, try the second.  */

    biosdrive = UseIOSYSDriveTables (dosdrive);
    if (biosdrive == 0xFF) biosdrive = UseInt13hTrap (dosdrive);
    return (biosdrive);

    #endif
}

/*  ------------------------------------------------------------------------  */

/*  First method.  This uses an undocumented int 2Fh function (available
    only since DOS 4) to obtain a pointer to the chain of tables in which
    IO.SYS maintains its information about the DOS drives supported by the
    default block device driver.  */

static struct IOSYSDRIVETABLE _far *GetIOSYSDriveTable (void);

static BYTE UseIOSYSDriveTables (WORD dosdrive)
{
    struct IOSYSDRIVETABLE _far *ptr;

    /*	Get the pointer.  */

    ptr = GetIOSYSDriveTable ();

    /*	Follow the chain (each table contains a far pointer to another - not
	necessarily in ascending drive order) to the table for the specified
	DOS drive number.  Another field in the table gives the BIOS drive
	number that is used for int 13h access.  */

    while ((WORD) ptr != 0xFFFF) {
	if (ptr -> cDOSDrive == (BYTE) dosdrive) return (ptr -> cBIOSDrive);
	ptr = ptr -> lpNext;
    }

    /*	If the end of the chain is reached without matching the DOS drive
	number, then the drive is not managed by the default block device
	driver in IO.SYS and the int 13h drive number remains unknown.	*/

    return (0xFF);
}

static struct IOSYSDRIVETABLE _far *GetIOSYSDriveTable (void)
{
    if (_osmajor < 4) {
	bad:
	return ((struct IOSYSDRIVETABLE _far *) -1);
    }
    _asm {
	    mov     ax,0800h
	    int     2Fh
	    cmp     al,0FFh
	    jnz     bad

	    push    ds
	    mov     ax,0803h
	    int     2Fh
	    mov     ax,di
	    mov     dx,ds
	    pop     ds
    }
}

/*  ------------------------------------------------------------------------  */

/*  Second method.  This generates a sector read from DOS level, having
    first inserted a small piece of code in the int 13h chain to observe
    whether or not int 13h is used to control the drive hardware.  */

static void _interrupt _far int13h (struct INTPACK);
static void (_interrupt _far *lpOrgInt13h) (void);

static BYTE buffer [0x0200];
static BYTE int13h_dl;
static BOOL int13h_read;

extern void _dos_getbreakflag (BYTE *);
extern void _dos_setbreakflag (BYTE);
static BYTE OrgBreakFlag;

static struct DISKIO {
    DWORD dwStart;
    WORD wCount;
    void _far *lpBuffer;
} diskio;

static BYTE UseInt13hTrap (WORD dosdrive)
{
    /*	Disable Ctrl-C checking to prevent termination in the unlikely
	event that a break is pending when the _dos_setvect function is
	called to unhook the int 13h probe.  */

    _dos_getbreakflag (&OrgBreakFlag);
    if (OrgBreakFlag) _dos_setbreakflag (OFF);

    /*	Install this function's int 13h hook, remembering to initialise the
	variable in which the hook routine is meant to record the value of
	register dl (the drive number) when int 13h is used.  */

    lpOrgInt13h = _dos_getvect (0x13);
    _dos_setvect (0x13, int13h);
    int13h_dl = 0xFF;
    int13h_read = FALSE;

    /*	Trigger an int 13h call by attempting to read a sector (logical
	sector 0) from the DOS drive.  The int 25h facility is the lowest
	level service that works with DOS drive definitions.  In a simple
	world, int 25h does little more than verify that the drive is valid
	at DOS level, then call the responsible device driver.	The success
	or failure of the read attempt is immaterial.

	Of course, int 25h may be intercepted before it reaches the DOS
	kernel.  Whether or not this produces multiple int 13h calls,
	possibly even to different drives, is for the int 13h probe to deal
	with.  One thing that may be reasonably well assumed is that no
	int 24h exceptions can be generated (which simplifies the defence to
	inhibiting breaks, as done above).  */

    diskio.dwStart = 0; 		// int 25h may need arguments in a
    diskio.wCount = 1;			// parameter block (typically for
    diskio.lpBuffer = buffer;		// DOS 4 and higher)

    _asm {
	    push    ds			// int 25h doesn't preserve
	    push    bp			// registers in earlier DOS versions
	    push    di
	    push    si

	    cmp     _osmajor,4		// DOS 4 and higher are certain to
	    jnb     newstyle		// admit input as parameter block

	    mov     bx,offset buffer	// For earlier DOS versions, try
	    mov     cx,1		// the register style of input
	    xor     dx,dx
	    mov     al,byte ptr dosdrive
	    int     25h
	    pop     bx			// remove interrupt flags from stack
	    jnc     done

	    cmp     ax,0207h		// error 0207h indicates input should
	    jnz     done		// be presented in parameter block

	newstyle:
	    mov     bx,offset diskio
	    mov     cx,0FFFFh
	    mov     al,byte ptr dosdrive
	    int     25h

	    popf			// int 25h always leaves flags
	done:
	    pop     si
	    pop     di
	    pop     bp
	    pop     ds
    }

    /*	Detach the int 13h probe.  This is the DOS call that could terminate
	the program pre-emptively were break-checking not disabled.
	Remember that when BREAK=ON, DOS checks the keyboard device (the
	most recently-loaded character device driver that had bit 0 set in
	its attribute word) and the stdin handle for a Ctrl-C character
	before proceeding with the body of the int 21h call.  */

    _dos_setvect (0x13, lpOrgInt13h);

    /*	Now that the int 13h probe is disconnected, return Ctrl-C processing
	to its previous condition.  */

    if (OrgBreakFlag) _dos_setbreakflag (OrgBreakFlag);
    return (int13h_dl);
}

static void _interrupt _far int13h (struct INTPACK r)
{
    /*	There is no telling how much int 13h activity will be generated by
	the device driver.  Worse, the test is exposed to interruption and a
	TSR may attempt some disk operation.  Worse still, int 25h may have
	been trapped by a cautious disk cache that flushes possibly huge
	amounts of material to possibly many different disk drives.

	A read request is expected and the test cannot be thought conclusive
	if none is trapped (at least, not without imposing assumptions about
	the sequence of int 13h operations that a device driver might use to
	prepare for reading - and even then, the possibility that int 25h
	has been hooked means that such preparations might be for some drive
	other than the one whose BIOS-level identity is sought).

	The possibility exists that more than one read request may be seen.
	If this happens, the resulting confusion must be untangled, but
	nothing about the expected read request is sure;  it cannot be
	supposed that logical sector 0 translates to physical sector 1,
	head 0, cylinder 0 and it cannot even be thought that the device
	driver will ask for only one sector (it may be caching consecutive
	sectors internally).

	The best that can be arranged is a defence against multiple reads
	(from different drives) occurring while the int 13h probe is
	installed.  */

    if (r.ah == _DISK_READ) {
	if (NOT int13h_read) {
	    int13h_read = TRUE;
	    int13h_dl = r.dl;
	}
	else if (int13h_dl != 0xFF) {
	    if (r.dl != int13h_dl) int13h_dl = 0xFF;
	}
    }

    /*	All requests must be passed onwards since it is unclear which ones
	the probe would be entitled to fail.  This has the unfortunate
	consequence of "hitting the disk" even when the sequence from
	int 25h to int 13h encounters no interference.	*/

    _chain_intr (lpOrgInt13h);
}

/*  ************************************************************************  */

