
/*  ************************************************************************  *
 *				   doslevel.c				      *
 *  ************************************************************************  */

#include    <stdlib.h>
#include    <string.h>

#include    <cds.h>
#include    <dpb.h>
#include    <sysvars.h>

/*  Two methods are given for finding a DPB, given a 0-based drive number.
    The first method uses a documented int 21h function, but "hits the
    disk", which is generally a disadvantage if only the drive properties
    are being investigated.  Worse, the function is unsuccessful if the disk
    drive is empty.  In essence, the second method follows the steps taken
    by the DOS kernel in the first method.  It avoids disk access, but is
    susceptible to future changes in the CDS structure (as happened with
    version 4).  */

#ifdef _USE_DOS_API_

struct DPB _far *GetDPB (WORD drive)
{
    /*	For any valid local drive, int 21h function 32h returns with ds:bx
	addressing the DPB for the corresponding "physical" DOS drive.  It
	is an old DOS function that indicates failure by al = FFh rather
	than a set carry flag.	*/

    _asm {
	    push    ds
	    mov     dx,drive
	    inc     dx		// DOS function requires 1-based drive number
	    mov     ah,32h
	    int     21h
	    mov     dx,ds
	    pop     ds
	    cmp     al,0FFh
	    jz	    bad
	    mov     ax,bx
	    jmp     done
	bad:
	    xor     ax,ax	// return NULL if no DPB exists for drive
	    cwd
	done:
    }
}

#else

static struct SYSVARS _far *GetSysvarsAddress (void);
static struct CDS _far *GetCDS (WORD);

/*  The objective of the function GetDPB is to find the DPB for the given
    0-based drive by taking the same steps as the DOS kernel for int 21h
    function 32h, but without having to "hit the disk".  */

struct DPB _far *GetDPB (WORD drive)
{
    struct CDS _far *cdsptr;

    /*	Begin by finding the CDS for the drive and checking that the drive
	is valid and local.  */

    cdsptr = GetCDS (drive);
    if (cdsptr == NULL) return (NULL);
    if (NOT (cdsptr -> wFlags & CDS_VALID)) return (NULL);
    if (cdsptr -> wFlags & CDS_REMOTE) return (NULL);

    /*	Read the DPB address from the CDS.  This method - used by the DOS
	kernel for almost all its drive dereferences - accounts for
	indirection at directory level and allows a DPB to be found for
	SUBSTed drives.  */

    return (cdsptr -> lpDPB);
}

static struct SYSVARS _far *GetSysvarsAddress (void)
{
    _asm {
	    mov     ah,52h
	    int     21h
	    mov     ax,bx
	    mov     dx,es
    }
}

static struct CDS _far *GetCDS (WORD drive)
{
    struct SYSVARS _far *sysvarsptr;
    struct CDS3 _far *cds3ptr;

    sysvarsptr = GetSysvarsAddress ();
    if ((BYTE) drive >= sysvarsptr -> cLogicalDrives) return (NULL);

    /*	Use the drive number to index the array of CDSs.  Allow for the
	change in structure that accompanied DOS 4.  */

    if (_osmajor >= 4) return (sysvarsptr -> lpFirstCDS + drive);
    else {
	cds3ptr = (struct CDS3 _far *) sysvarsptr -> lpFirstCDS + drive;
	return ((struct CDS _far *) cds3ptr);
    }
}

#if FALSE

/*  An alternative way to find the DPB for a drive - but only if the drive
    corresponds directly to a device unit - is to follow the DPB chain.
    This is included here for reference only.  DOS itself does not use this
    method, except in a few cases where it is interested only in "physical"
    drives.  */

static struct DPB _far *GetRawDPB (WORD drive)
{
    struct DPB _far *ptr = GetSysvarsAddress () -> lpFirstDPB;

    while ((WORD) ptr != 0xFFFF) {
	if (ptr -> cDrive == (BYTE) drive) return (ptr);
	ptr = ptr -> lpNext;
    }
    return (NULL);
}

#endif
#endif

/*  ========================================================================  */

/*  Similarly, two methods are provided for analysing properties lent to a
    drive via the CDS.	Again, using the API (undocumented in this case)
    triggers a disk access.

    For both methods, pass the results as a set of printf style arguments
    ready for output via a function called PutComment.	This lets an
    external module ensure consistent formatting.  */

extern void PutComment (CHAR *, ...);

static CHAR pathname [0x80];

#ifndef _USE_DOS_API_

void DescribeCDSProperties (WORD drive)
{
    struct CDS _far *cdsptr;
    register WORD count;
    register CHAR *ptr;

    cdsptr = GetCDS (drive);
    if (cdsptr == NULL OR NOT (cdsptr -> wFlags & CDS_VALID)) return;

    if (cdsptr -> wFlags & (CDS_SUBST | CDS_REMOTE)) {
	count = cdsptr -> wRootCount;
	_fmemcpy (pathname, cdsptr -> cDirectory, count);
	ptr = pathname + count;
	if (count == 2) *ptr ++ = '\\';
	*ptr = '\0';
	PutComment ("%s %s",
	    cdsptr -> wFlags & CDS_SUBST ? "substituted for" : "mapped to",
	    pathname);
    }

    if (cdsptr -> wFlags & (CDS_JOIN))
	PutComment ("joined to %Fs", cdsptr -> cDirectory);
}

#else

static WORD _dos_canonicalise (CHAR *, CHAR *);

#define NULCHARS    (sizeof ("\\NUL") - 1)

#pragma pack (1)
static struct {
    CHAR drive;
    CHAR colon;
    CHAR backslash_nul [NULCHARS + 1];
} test_spec = {'X', ':', "\\NUL"};
#pragma pack ()

void DescribeCDSProperties (WORD drive)
{
    WORD parsedlength;

    /*	One way to see if a drive is mapped to another drive's directory
	tree is to use int 21h function 60h (undocumented, but indispensable
	to many standard DOS utilities and raised to the level of the DOS
	command line as TRUENAME) to resolve a drive's root directory
	(letter, colon and backslash).	However, this will ordinarily "hit
	the disk".

	An exceptional case may be used to advantage.  Instead of parsing
	just the drivespec and leading backslash, parse a device name in the
	root directory.  This will cause DOS to consult the CDS to validate
	the drive, but shortly afterwards, DOS will exit from its parsing
	routines when it discovers that the "file" is really just a
	character device.  This complicates the programming a little, since
	the name of the character device has to be removed from the parsed
	result before the drive translation can be reported - but it avoids
	"hitting the disk".

	One detraction that seems unavoidable is that short of using trial
	and error on all the subdirectories of every drive on the system,
	int 21h function 60h will not reveal the directory to which a drive
	has been JOINed.  */

    test_spec.drive = (BYTE) (drive + 'A');
    if (_dos_canonicalise (pathname, (CHAR *) &test_spec) != 0) return;

    /*	If the input path changes when canonicalised, then the drive is
	mapped elsewhere in the file system.  Of course, the \NUL component
	at the end of the path must be removed in order to find out what the
	bare drivespec translates to.  */

    if (memcmp (pathname, (CHAR *) &test_spec, sizeof (test_spec)) != 0) {
	parsedlength = strlen (pathname);
	if (parsedlength > NULCHARS) {
	    if (memcmp (pathname + parsedlength - NULCHARS,
		    test_spec.backslash_nul, NULCHARS) == 0) {
		*(pathname + parsedlength - NULCHARS) = '\0';
		PutComment ("translates to %s", pathname);
	    }
	}
    }
}

static WORD _dos_canonicalise (CHAR *outptr, CHAR *inptr)
{
    _asm {
	    mov     si,inptr
	    push    ds
	    pop     es
	    mov     di,outptr
	    mov     ah,60h
	    int     21h
	    jc	    done
	    xor     ax,ax
	done:
    }
}

#endif

/*  ************************************************************************  */

