
/*  ************************************************************************  *
 *		  addfiles.c - primary module for addfiles.exe		      *
 *  ************************************************************************  */

#include    <dos.h>
#include    <memory.h>
#include    <process.h>
#include    <signal.h>
#include    <stdarg.h>
#include    <stdio.h>
#include    <stdlib.h>
#include    <string.h>

#include    <arena.h>
#include    <psp.h>
#include    <sft.h>
#include    <standard.h>

#pragma intrinsic (_disable, _enable)

static char description [] = "\n\
Adds a specified number of System File Tables (SFTs).\n\
\n\
  ADDFILES [ExtraSFTs] [/u]\n\
\n\
If the number of new SFTs is not given, ADDFILES reports the current count.\n\
\n\
Under DOS 5 and higher, space for the new SFTs will be sought in upper\n\
memory first.  If this fails, the search continue in conventional memory\n\
by default.  Use the /u switch to insist on trying upper memory only.\n";

/*  Functions and data to control memory allocation  */

extern WORD _dos_getstrategy (WORD *);		    // in memstrat.c
extern WORD _dos_getumblink (BOOL *);		    // in memstrat.c
extern WORD _dos_setstrategy (WORD);		    // in memstrat.c
extern WORD _dos_setumblink (BOOL);		    // in memstrat.c

WORD wOrgStrategy;
BOOL bOrgUMBLink;
BOOL bDemandUpper = FALSE;	    // set to TRUE if /u switch present

/*  Functions and data for SFT location and construction  */

extern struct SFT_HEADER _far *GetFirstSFTHeader (void);    // in sft.c
extern WORD GetSFTCount (void); 			    // in sft.c
extern WORD GetSFTSize (void);				    // in sft.c

void BuildNewSFTs (_segment, WORD, WORD);
void AddSFTBlock (struct SFT_HEADER _far *, WORD);

WORD wAllocParas, wNewSFTs = 0;

/*  Functions and data concerned with reloading program into higher memory
    to make way for new SFTs  */

BOOL bReloaded;
_segment spParentCode, spParentData, spParentPSP;

BOOL HaveBeenReloaded (void);
void KillParent (void);

/*  Exit handling  */

int _far HardErrorTrap (void);
void cleanup (void);

void PutError (char *, ...);

/*  ========================================================================  */

int main (int argc, char **argv)
{
    char *progname;
    register char *ptr;
    BYTE ch;
    WORD sftcount, sftsize, sftblockbytes;
    register WORD errorcode;
    WORD allocseg;

    if (MK_WORD (_osmajor, _osminor) < MK_WORD (3, 10)) {
	PutError ("requires DOS version 3.10 or higher");
	return (-1);
    }

    /*	It simplifies matters greatly if premature termination via the
	int 23h and int 24h DOS exceptions can be discounted.  */

    signal (SIGINT, SIG_IGN);
    _harderr (HardErrorTrap);

    /*	The program may be reloaded with a LAST_FIT strategy in order to
	move it out the way.  If this is the first instance, then parse the
	command line and determine whether or not the new SFTs can be added
	immediately.  If this is the reload, perform the necessary memory
	juggling and then build the new SFTs.  */

    bReloaded = HaveBeenReloaded ();
    if (NOT bReloaded) {

	/*  Process the command line, but save the program name for later
	    in case reloading is required.  */

	progname = *argv;
	while (++ argv, -- argc) {
	    ptr = *argv;
	    if (*ptr == '/') {
		ch = ptr [1] | (BYTE) 0x20;
		if (ptr [2] == '\0') {
		    if (ch == '?') {
			printf (description);
			return (0);
		    }
		    if (ch == 'u') {
			bDemandUpper = TRUE;
			continue;
		    }
		}
		PutError ("invalid switch %s", ptr);
		return (-1);
	    }
	    else {
		if (wNewSFTs != 0) {
		    PutError ("too many parameters");
		    return (-1);
		}
		wNewSFTs = atoi (ptr);
		if (wNewSFTs == 0) {
		    PutError ("invalid parameter %s", ptr);
		    return (-1);
		}
	    }
	}

	/*  The current SFT count is required anyway - in case the proposed
	    addition would pass the fixed limit of 255 - so use it for a
	    little report if no command line parameters were detected.	*/

	sftcount = GetSFTCount ();
	if (wNewSFTs == 0) {
	    printf ("\nThere are presently %d SFTs\n", sftcount);
	    return (0);
	}
	if (sftcount + wNewSFTs > 0xFF) {
	    PutError ("total would exceed 255 - can add only %d new SFTs",
			0xFF - sftcount);
	    return (-1);
	}

	/*  Calculate the size of the external memory allocation in which
	    the new SFTs will be built.  Include a paragraph for a memory
	    arena subheader under DOS 4.0 and higher.  */

	sftsize = GetSFTSize ();
	if (sftsize == 0) {
	    PutError ("cannot determine size of System File Table");
	    return (-1);
	}
	sftblockbytes = sizeof (struct SFT_HEADER) + sftsize * wNewSFTs;
	wAllocParas = (sftblockbytes + 15) >> 4;
	if (_osmajor >= 4) wAllocParas ++;

	/*  Now that the preliminaries are over, obtain the current memory
	    allocation strategy (and UMB link in DOS 5 and higher), then
	    install a C run-time terminator to ensure restoration when the
	    program exits.  */

	_dos_getstrategy (&wOrgStrategy);
	if (_osmajor >= 5) _dos_getumblink (&bOrgUMBLink);

	atexit (cleanup);

	/*  In DOS 5 and higher, attempt first to find upper memory for the
	    new SFTs.  This is the easiest case.  There is an assumption
	    here that the program is being run from conventional memory, so
	    the danger of fragmentation does not arise - in the present
	    implementation therefore, the program should not itself be
	    loaded high.  */

	if (_osmajor >= 5) {
	    _dos_setstrategy (wOrgStrategy | HIGHONLY);
	    errorcode = _dos_setumblink (ON);
	    if (errorcode == 0) {
		errorcode = _dos_allocmem (wAllocParas, &allocseg);
		if (errorcode == 0) {

		    /*	The BuildNewSFTs function is expected to arrange the
			retention of the memory used for the new SFTs - a
			natural consequence of emulating the usual
			construction of SFTs in the SYSINIT configuration
			block.	The program itself does not stay resident.  */

		    BuildNewSFTs (allocseg, wAllocParas, wNewSFTs);
		    return (wNewSFTs);
		}
	    }
	    if (bDemandUpper) {
		PutError ("insufficient free upper memory");
		return (-1);
	    }
	}

	/*  An external allocation for the new SFTs would probably result in
	    memory fragmentation.  Neither is it particularly desirable to
	    retain the 256 bytes of the PSP - which would happen if the
	    program finished with the TSR exit, having built the new SFTs
	    over space presently occupied by the program's code.  Instead,
	    the aim is to reload the program elsewhere and have the second
	    copy (higher in memory) terminate the first, thereby securing
	    its memory and avoiding fragmentation.  */

	_dos_setstrategy (LAST_FIT);
	if (_osmajor >= 5) _dos_setumblink (OFF);

	errorcode = spawnl (P_WAIT, progname, progname, NULL);

	PutError ("unable to reload into higher memory (%s)",
	    errorcode == -1 ? strerror (errno) : "internal error");
	return (-1);
    }
    else {

	/*  This block executes because the program has been reloaded.	Two
	    parameters describing the required allocation must be retrieved
	    from the first instance of the program.  */

	wAllocParas = *((WORD _far *) MK_FP (spParentData, &wAllocParas));
	wNewSFTs = *((WORD _far *) MK_FP (spParentData, &wNewSFTs));

	/*  Take over from the first instance of the program, then allocate
	    memory and build the new SFTs just as if the reload never
	    happened.  The parent's termination, incidentally, causes the
	    memory allocation strategy to be restored, so there is no need
	    for special exit handling in this reload case.  */

	KillParent ();
	errorcode = _dos_allocmem (wAllocParas, &allocseg);
	if (errorcode != 0) {
	    PutError ("error allocating memory after relocation");
	    return (-1);
	}
	BuildNewSFTs (allocseg, wAllocParas, wNewSFTs);
	return (wNewSFTs);
    }
}

void BuildNewSFTs (_segment allocseg, WORD paras, WORD wNewSFTs)
{
    _segment arenaseg, sftblockseg, subarenaseg;

    /*	Clear the memory block.  */

    _fmemset (MK_FP (allocseg, NULL), 0, paras << 4);

    #define arenaptr(seg) ((struct ARENA_HEADER _far *) MK_FP (seg, NULL))
    #define sftheaderptr(seg) ((struct SFT_HEADER _far *) MK_FP (seg, NULL))

    /*	If the DOS version is 4.0 or higher, the first paragraph of the
	allocation is intended as a memory arena subheader - to emulate the
	way that SYSINIT builds its SFT block.	The MEM program may use this
	construction to identify the new SFTs as an extra FILES= entry.  */

    if (_osmajor >= 4) {
	sftblockseg = allocseg + 1;
	subarenaseg = allocseg;
	arenaptr (subarenaseg) -> cSignature = 'F';
	arenaptr (subarenaseg) -> spOwner = sftblockseg;
	arenaptr (subarenaseg) -> wParagraphs = paras - 1;
    }
    else sftblockseg = allocseg;

    /*	Append the new SFT block to the existing chain.  */

    AddSFTBlock (MK_FP (sftblockseg, NULL), wNewSFTs);

    /*	Assign the memory arena header as if it had been obtained for
	configuration during system startup.  This reassignment also serves
	as the way to keep the block after the program terminates.  */

    arenaseg = allocseg - 1;
    arenaptr (arenaseg) -> cName [0] = 'S';
    arenaptr (arenaseg) -> cName [1] = 'D';
    arenaptr (arenaseg) -> spOwner = ARENA_DOS;
}

void AddSFTBlock (struct SFT_HEADER _far *newsftheader, WORD wNewSFTs)
{
    struct SFT_HEADER _far *ptr;

    /*	Initialise the header for the new SFTs.  */

    newsftheader -> lpNext = (struct SFT_HEADER _far *) -1;
    newsftheader -> wCount = wNewSFTs;

    /*	Follow the chain of SFT headers to its end, then patch in the
	address of the new SFTs.  Remember to disable interrupts while
	loading the far pointer.  */

    ptr = GetFirstSFTHeader ();
    while (*((WORD _far *) ptr) != -1) {
	ptr = ptr -> lpNext;
    }
    _disable ();
    ptr -> lpNext = newsftheader;
    _enable ();
}

int _far HardErrorTrap (void)
{
    return (_HARDERR_FAIL);
}

void cleanup (void)
{
    _dos_setstrategy (wOrgStrategy);
    if (_osmajor >= 5) _dos_setumblink (bOrgUMBLink);
}

void PutError (char *str, ...)
{
    va_list va_start (argptr, str);
    printf ("\nADDFILES fatal error:  ");
    vprintf (str, argptr);
    printf ("\n");
    va_end (argptr);
}

/*  ========================================================================  */

/*  The next function is concerned with identifying whether this program is
    being re-run by an earlier copy of itself in order to move out of the
    way for the new SFTs.  */

static char idstring [] = "ADDFILES";

BOOL HaveBeenReloaded (void)
{
    #define pspptr(seg) ((struct PSP _far *) MK_FP (seg, NULL))

    /*	Inspect this program's PSP to find the parent's PSP.  Work on the
	assumption that they are just different instances of the same
	program.  Record the parent's presumed code and data segments in
	global variables for later use, then check that the ID string is at
	the same offset in both the parent's and this program's DGROUP.  */

    spParentPSP = pspptr (_psp) -> spParentPSP;
    spParentCode = spParentPSP + _segname ("_CODE") - _psp;
    spParentData = spParentPSP + _segname ("_DATA") - _psp;
    if (_fmemcmp (MK_FP (spParentData, &idstring), &idstring,
	    sizeof (idstring)) == 0) return (TRUE);
    else return (FALSE);
}

/*  Some highly compiler-dependent unpleasantries - explained below  */

typedef void _interrupt _far INTERRUPT_HANDLER (void);

extern INTERRUPT_HANDLER _far *_aintdiv;
extern INTERRUPT_HANDLER _cintDIV;

void KillParent (void)
{
    void _far *stackmarker;
    void (_far *retaddr) (void);

    /*	Acquire the parent's termination "state" - this program will be
	going back to its parent's parent.  */

    #if FALSE
    pspptr (_psp) -> lpInt22h = pspptr (spParentPSP) -> lpInt22h;
    pspptr (_psp) -> lpInt23h = pspptr (spParentPSP) -> lpInt23h;
    pspptr (_psp) -> lpInt24h = pspptr (spParentPSP) -> lpInt24h;
    pspptr (_psp) -> spParentPSP = pspptr (spParentPSP) -> spParentPSP;
    #else
    struct TERM_STATE { 			// very efficient method of
	INTERRUPT_HANDLER _far *lpInt22h;	// copying consecutive
	INTERRUPT_HANDLER _far *lpInt23h;	// fields in structure -
	INTERRUPT_HANDLER _far *lpInt24h;	// compiler can't possibly
	_segment spParentPSP;			// miss the optimisation
    };
    #define termstateptr(seg)	((struct TERM_STATE _far *) \
	    &(((struct PSP _far *) ((DWORD) (seg) << 16)) -> lpInt22h))
    *termstateptr (_psp) = *termstateptr (spParentPSP);
    #endif

    /*	Skipping a generation invalidates the int 00h vector recorded by the
	C run-time.  The compiler dependence introduced to overcome this is
	best viewed as comeuppance for interfering with another process
	while having only incomplete control of one's own program.  As an
	alternative to using the run-time libraries particular labels,
	consider trapping the parent's restoration of the int 00h vector.  */

    #define lpParents_aintdiv	((INTERRUPT_HANDLER _far * _far *) \
					MK_FP (spParentData, &_aintdiv))
    _aintdiv = *lpParents_aintdiv;
    *lpParents_aintdiv = _cintDIV;

    /*	Set a new state for the parent - to return further down this
	procedure when it terminates.  */

    _asm {						// in-line assembler
	    mov     word ptr retaddr,offset resume	// needed to resolve
	    mov     word ptr retaddr+2,cs		// forward reference
    }							// to code label
    pspptr (spParentPSP) -> lpInt22h = retaddr;
    pspptr (spParentPSP) -> lpInt23h = _dos_getvect (0x23);
    pspptr (spParentPSP) -> lpInt24h = _dos_getvect (0x24);
    pspptr (spParentPSP) -> spParentPSP = _psp;

    /*	Now that the parent-child relationship has been reversed, the stack
	adopted when the parent terminates will be taken from the relevant
	field in this program's PSP.  Build a stack with registers in the
	order expected by the DOS kernel and load the address into the PSP.  */

    _asm {
	    xor     ax,ax
	    push    ax		// space for flags and return address -
	    push    ax
            push    ax

	    push    es		// registers to be recovered at end of
	    push    ds		// function
	    push    bp
	    push    di
	    push    si
	    push    dx
	    push    cx
	    push    bx
	    push    ax

	    mov     word ptr stackmarker,sp
	    mov     word ptr stackmarker+2,ss
    }
    pspptr (_psp) -> lpStack = stackmarker;

    /*	Make the parent the active process and terminate it.  Ordinarily,
	this would mean calling int 21h function 4Ch, but such pre-emption
	is unnecessary here because the parent is known;  indeed, it is
	another copy of this same program, so it is possible to call the
	exit (0) function in the parent's code segment and allow it the
	chance to clean up properly.  */

    _asm {
	    mov     bx,spParentPSP
	    mov     ah,50h
	    int     21h

	    xor     ax,ax		// argument on stack for exit ()
	    push    ax
	    mov     ax,offset exit	// place segment and offset of
	    push    spParentCode	// parent's exit function on stack
	    push    ax			// ready for transfer via RETF

	    mov     ax,spParentData	// load ds with parent't DGROUP
	    mov     ds,ax

	    retf			// go to parent - the earlier
	resume: 			// manipulations bring control
    }					// back here
}

/*  ************************************************************************  */

