
/*  ************************************************************************  *
 *		 eatblock.c - top-level module for EATBLOCK.EXE 	      *
 *  ************************************************************************  */

#include    <dos.h>
#include    <errno.h>
#include    <process.h>
#include    <stdarg.h>
#include    <stdio.h>
#include    <stdlib.h>
#include    <string.h>

#include    <arena.h>
#include    <doserror.h>
#include    <standard.h>

char description [] = "\n\
Allocates a memory block specified by both size and location.\n\
\n\
  EATBLOCK size [base [ceiling]] [/p]\n\
\n\
All numeric values are interpreted as hexadecimal.  Size is in paragraphs.\n\
Base and ceiling are segment addresses.\n\
\n\
Use size=FFFF to report on the largest free block within the given range.\n\
\n\
Ordinarily, EATBLOCK starts a command processor to support testing under the\n\
new memory environment.  The memory block is released when the command\n\
processor exits.  The /p switch instructs EATBLOCK to make its allocation\n\
permanent and to exit immediately (i.e., without starting a shell).\n";

/*  Subfunctions to separate these cases (with and without /p, respectively)
    and a string for use when starting a command shell	*/

static int AssignBlock (_segment);
static int RunTestShell (void);

static char instructions [] = "\n\
A command processor is being started to allow investigation with\n\
whatever tools you prefer.  Type EXIT when satisfied.\n";

/*  Special memory allocation function - similar to the Microsoft C library
    function _dos_allocmem, but with extra arguments to specify range and
    passing separate pointers for base and size results.  The function is
    defined and annotated in the file MALLOC.C.  */

extern WORD _dos_special_allocmem (WORD, _segment, _segment,
					    _segment *, WORD *);
/*  Miscellany  */

#define arenaptr(seg)	((struct ARENA_HEADER _far *) MK_FP (seg, NULL))

extern BOOL ReadHexWord (char *, WORD *);	    // in readhex.c
static void PutError (char *, ...);

/*  ========================================================================  */

int main (int argc, char **argv)
{
    char *ptr;
    WORD ch, errorcode, size, temp;
    _segment base = 0x0000, ceiling = 0xFFFF;
    BOOL sizedone = FALSE, basedone = FALSE, ceilingdone = FALSE;
    BOOL permanent = FALSE;

    /*	Parse the command line.  */

    while (++ argv, -- argc) {
	ptr = *argv;
	if (*ptr == '/') {
	    if (ptr [2] == '\0') {
		ch = ptr [1] | 0x20;
		if (ch == '?') {
		    printf (description);
		    return (0);
		}
		else if (ch == 'p') {
		    permanent = TRUE;
		    continue;
		}
	    }
	    PutError ("invalid switch %s", ptr);
	    return (-1);
	}
	else {
	    if (NOT ReadHexWord (ptr, &temp)) {
		PutError ("invalid parameter %s", ptr);
		return (-1);
	    }
	    if (NOT sizedone) {
		size = temp;
		sizedone = TRUE;
	    }
	    else if (NOT basedone) {
		base = temp;
		basedone = TRUE;
	    }
	    else if (NOT ceilingdone) {
		ceiling = temp;
		ceilingdone = TRUE;
	    }
	    else {
		PutError ("too many parameters %s", ptr);
		return (-1);
	    }
	}
    }
    if (NOT sizedone) {
	printf (description);
	return (0);
    }

    /*	Attempt to allocate a memory block of the desired size within the
	specified range.  */

    errorcode = _dos_special_allocmem (size, base, ceiling, &base, &size);

    if (errorcode != 0) {

	/*  All forms of allocation error are terminal.  */

	if (errorcode == E_OUT_OF_MEMORY) {

	    /*	The usual reason for failure is simply that no free memory
		block meets the requirements, in which case, report the size
		of the largest suitable block.	*/

	    printf ("\nLargest block available between segments %04Xh "
		    "and %04Xh is %04Xh paragraphs\n", base, ceiling, size);
	}
	else if (errorcode == E_ACCESS_DENIED) {

	    /*	Another cause is that the _dos_special_allocmem function
		found that it could not hold information about memory blocks
		outside the requested range while forcing DOS to allocate
		from within the range.	*/

	    PutError ("too much memory fragmentation, too little heap space");
	}
	else {
	    /*	Any other error condition almost certainly indicates that
		the memory control chain is corrupt.  */

	    PutError ("serious memory control problem");
	}
	return (-1);
    }
    /*	Having "eaten" a suitable block of memory, affirm the success, then
	either assign the block permanently and exit or locate the command
	processor and run it as a test shell.  */

    printf ("\nAllocated %04Xh paragraphs at segment %04Xh\n", size, base);
    return (permanent ? AssignBlock (base) : RunTestShell ());
}

static int RunTestShell (void)
{
    register char *pCOMSPEC;
    int exitcode;

    /*	Consult the environment for the command processor's pathname,
	display an explanatory message and then run the shell (with an
	empty command line).  */

    pCOMSPEC = getenv ("COMSPEC");
    if (pCOMSPEC == NULL) {
	PutError ("cannot find command processor for test");
	return (-1);
    }
    printf (instructions);
    exitcode = spawnl (P_WAIT, pCOMSPEC, pCOMSPEC, NULL);

    if (exitcode == -1) PutError ("%s", strerror (errno));
    else {
	/*  On leaving the command processor, remind the user that the shell
	    was executed as part of a memory allocation test - just for
	    clarity, it being far too easy to get lost when multiple copies
	    of the command processor are run.  */

	printf ("\nEATBLOCK:  memory allocation test completed.\n");
    }
    return (exitcode);
}

#pragma optimize ("a", on)

static int AssignBlock (_segment base)
{
    /*	Ordinarily, the block will be released when the program terminates
	(without intending to stay resident).  This may be prevented by
	reassigning the block's ownership beforehand.

	Incidentally, without the /Oa option (forced by the optimize pragma
	preceding the function), the compiler has little choice but to
	recalculate base - 1 in each of the following lines, since no matter
	how improbably, the assignments could conceivably alter the value of
	base from line to line.  */

    arenaptr (base - 1) -> cName [0] = 'S';             // system
    arenaptr (base - 1) -> cName [1] = 'C';             // code
    arenaptr (base - 1) -> spOwner = ARENA_DOS; 	// 0008h

    return (0);
}

#pragma optimize ("a", )

static void PutError (char *str, ...)
{
    va_list va_start (argptr, str);
    printf ("\nEATBLOCK fatal error:  ");
    vprintf (str, argptr);
    printf ("\n");
    va_end (argptr);
}

/*  ************************************************************************  */

