
/*  ************************************************************************  *
 *				   dospawn.c				      *
 *  ************************************************************************  */

/*  This file provides a generalisation of DOS process management to allow
    specification of the memory region into which a program is loaded for
    execution.	*/

#include    <dos.h>
#include    <errno.h>
#include    <process.h>
#include    <stdlib.h>

#include    <exec.h>
#include    <fcb.h>
#include    <intpack.h>
#include    <psp.h>
#include    <standard.h>

#if NOT (defined (M_I86SM) OR defined (M_I86CM))
#error	prepared only for memory models with small data size
#endif

/*  Functions defined in MEMBLOCK.C to force memory allocations into a
    given region by masking off all other memory  */

extern WORD MaskMemoryRegion (_segment, _segment, WORD *);
extern WORD ReleaseMask (void *);

/*  Parameters defining the region of interest	*/

_segment spBase = 0x0000, spCeiling = 0xFFFF;

/*  Parameters selecting a memory environment for the child program - only
    the upper memory link is configurable in this demonstration  */

WORD wChildsStrategy;
BOOL bChildsUMBLink;

BOOL bSetChildsStrategy = FALSE, bSetChildsUMBLink = FALSE;

BOOL bRestrictChildsMemory = FALSE;

/*  Functions and data to bring upper memory into the scheme  */

extern WORD _dos_getstrategy (WORD *);		    // in memstrat.c
extern WORD _dos_getumblink (BOOL *);		    // in memstrat.c
extern WORD _dos_setstrategy (WORD);		    // in memstrat.c
extern WORD _dos_setumblink (BOOL);		    // in memstrat.c

WORD wOrgStrategy;
BOOL bOrgUMBLink;

/*  Functions and data to control possible termination by Ctrl-C or
    Ctrl-Break while the upper memory link has been changed  */

extern void _dos_getbreakflag (BYTE *); 	    // in break.c
extern void _dos_setbreakflag (BYTE);		    // in break.c

BYTE bOrgBreakFlag;
void (_interrupt _far *lpOrgInt24h) (void);

void _interrupt _far int24h (struct INTPACK r)
{
    r.al = _HARDERR_FAIL;
}

extern BOOL _child;	    // for the signal function's int 23h handler

/*  Macro to ease PSP management  */

#define PSP_FP(seg) ((struct PSP _far *) MK_FP (seg, NULL))

/*  Data structures required for DOS function 4B01h  */

static struct FCB fcb1;
static struct FCB fcb2;
static struct LOADBLOCK loadblock;

/*  Data structure required for DOS function 4B05h in DOS 5  */

static struct EXECSTATE execstate;

/*  Where possible, assembly language code is placed in separate C language
    functions.	*/

static WORD GetChildsExitCode (void);
static _segment GetCurrentProcess (void);
static void InitFCBs (char *);
static void SetExecState (void);

/*  Prototype for assembly language routine in Microsoft C library which
    maps DOS error numbers into the scheme presented in the ERRNO.H file.  */

extern int _dosretax (void);

/*  Note that _dosretax is a common tail end for the many library functions
    that must load error values into the errno global variable.  It is not
    a function in its own right but more like a dislocated piece of the
    function that jumps to it.

    Functions jump to _dosretax with a clear carry flag if they mean to
    return the value in ax as their own function value.  Alternatively,
    they jump with a set carry flag if they mean to return -1 as an error
    indication (like the family of spawn functions), but need to have the
    value in ax (a DOS error) be converted and recorded.

    The C prototype given above is therefore strictly incorrect.  It is
    needed however as an external reference for the following brief
    function which bridges the gap between the assembly language and the
    use of _dosretax by C functions outside the Microsoft C library.  */

static int dosretax (WORD value_for_ax, BOOL value_for_cf)
{
    _asm {
	    xor     ax,ax
	    sub     ax,value_for_cf
	    mov     ax,value_for_ax
	    call    _dosretax		// _dosretax "returns" directly to
    }					// the function that called dosretax
}

/*  The main function in this module replaces an internal component at the
    heart of all spawn functions in the Microsoft C library.  */

int _dospawn (int mode, char *progname, char *cmdline, char *env)
{
    register WORD exitcode;
    BOOL convert_exitcode;
    WORD maskhandle;
    register _segment childpsp;
    BOOL execution_attempted, first_time_through;

    /*	Check the spawn mode.  Under DOS, the Microsoft C library supports
	only the value P_WAIT defined in the PROCESS.H file (not counting
	overlays). If the function has passed an unsupported mode, then -1
	should be returned and the errno global variable should be loaded
	with an appropriate value from the list in ERRNO.H.  */

    if (mode != P_WAIT) {
	errno = EINVAL;
	exitcode = 0xFFFF;
	convert_exitcode = FALSE;
    }
    else {

	/*  Until the child program has actually been executed, all further
	    values loaded into exitcode will be DOS error numbers requiring
	    last minute conversion into the errno scheme.  */

	convert_exitcode = TRUE;

	/*  Load the environment and command line addresses into the
	    parameter block for the DOS "load program" function.  */

	loadblock.spEnvironment = (_segment) env + ((WORD) env >> 4);
	loadblock.lpCommandLine = cmdline;
	loadblock.lpFCB1 = &fcb1;
	loadblock.lpFCB2 = &fcb2;
	InitFCBs (cmdline);

	if (_osmajor >= 5) {
	    _dos_getstrategy (&wOrgStrategy);
	    _dos_getumblink (&bOrgUMBLink);
	}

	/*  Bringing upper memory into play while loading the child may
	    involve changing the upper memory link.  If so, steps must also
	    be taken to ensure that the parent program cannot terminate (via
	    int 23h or 24h exceptions) until the link has been restored to
	    its original condition.  It is easiest just to prevent these
	    exceptions from either occurring or causing any problem.  Since
	    the code involved will be using only those DOS functions which
	    do not check Ctrl-C and Ctrl-Break unless the break flag is on,
	    it suffices just to turn that flag off.  Critical errors may
	    occur during DOS function 4Bh (used to load the child program
	    from its file on disk):  they must be intercepted and returned
	    harmlessly.  These complications do not arise if the upper
	    memory link is already enabled.  */

	if (_osmajor >= 5) {
	    _dos_getumblink (&bOrgUMBLink);
	    if (NOT bOrgUMBLink) {
		_dos_getbreakflag (&bOrgBreakFlag);
		if (bOrgBreakFlag) _dos_setbreakflag (OFF);
		lpOrgInt24h = _dos_getvect (0x24);
		_dos_setvect (0x24, int24h);
		_dos_setumblink (ON);
	    }
	}

	/*  To force the child program into the required memory region,
	    allocate all other memory temporarily - as a sort of mask.	*/

	exitcode = MaskMemoryRegion (spBase, spCeiling, &maskhandle);

	/*  Unless the memory masking failed, load the program.  Only the
	    memory between the designated base and ceiling segments will be
	    available for the program's environment and main allocation.
	    Use DOS function 4B01h so that execution of the child's code can
	    be delayed until the memory mask has been released.  This means
	    the child program can be forced to load in a particular area of
	    memory without affecting its subsequent memory usage (or that of
	    other programs).

	    Observe that when DOS function 4B01h initialises the child, it
	    cannot know anything of the circumstances under which the
	    child's code may eventually be executed.  The parent and child
	    PSPs are prepared with values that apply at the time DOS
	    function 4B01h is called.  This means that the instruction
	    labelled termaddr in the following code will be executed not
	    only as the first instruction after the child has been loaded,
	    but again when the child terminates.  */

	if (exitcode == 0) {

	    first_time_through = TRUE;
	    execution_attempted = FALSE;

	    _asm {
		    push    ds
		    pop     es
		    mov     bx,offset loadblock
		    mov     dx,progname
		    mov     ax,4B01h
		    int     21h
		termaddr:
		    jc	    done
		    xor     ax,ax
		done:
		    mov     exitcode,ax
	    }
	    if (execution_attempted) _child = FALSE;
	}

	if (exitcode != 0) {

	    /*	If the child has not been loaded successfully, then only a
		little cleaning up remains to be done.	In particular, if
		the upper memory link was changed, then it must be restored.
		In this case, the break flag must also be returned to its
		original condition, as must the temporary int 24h handling.
		For balance, the memory mask (if allocated) is also
		released, though of course, DOS will take care of this
		anyway when the program exits (soon).  */

	    if (_osmajor >= 5) {
		if (NOT bOrgUMBLink) {
		    _dos_setumblink (bOrgUMBLink);
		    _dos_setvect (0x24, lpOrgInt24h);
		    _dos_setbreakflag (bOrgBreakFlag);
		}
	    }
	    ReleaseMask (&maskhandle);
	}
	else if (first_time_through) {
	    first_time_through = FALSE;

	    /*	The absence of an error code from DOS function 4B01h means
		that the child process has been initialised successfully.
		Moreover, it will in fact be the current process (even
		though its own code loaded from disk may not be executed
		for some time yet).  */

	    childpsp = GetCurrentProcess ();

	    /*	A high priority must be given to ensuring that the child's
		eventual termination will be safe.  The question of where
		control will resume has already been handled, but there
		remains some difficulty with restoration of this function's
		stack and registers.  This must certainly be dealt with
		before the child has any opportunity to terminate.

		Ordinarily, a DOS caller's registers are saved on the
		caller's stack and the resulting SS:SP is entered into the
		DOS caller's PSP.  For DOS function 4B00h, the caller
		becomes a parent;  DOS switches stacks to one provided by
		the child program, with the result that the parent's stack
		and registers are preserved until the child terminates.

		DOS function 4B01h pops the stacked registers and returns -
		just like any DOS function.  The stack pointer loaded into
		the PSP remains however, so correct alignment upon the
		child's termination can be arranged by rebuilding the stack
		created when DOS function 4B01h was entered.  */

	    _asm {
		    xor     ax,ax	// space for three words
		    push    ax		// corresponding to the
		    push    ax		// interrupt return address
		    push    ax		// and flags

		    push    es		// registers in same order
		    push    ds		// as usually saved by DOS
		    push    bp		// kernel when dispatching
		    push    di		// int 21h functions
		    push    si
		    push    dx
		    push    cx
		    push    bx
		    push    ax		// this brings the stack down
	    }				// to the restoration address
					// recorded in the parent PSP

	    /*	The same cleaning up as before is required, but there is an
		additional subtlety regarding the temporary int 24h
		handling, for the address of this function's temporary
		int 24h handler will also have been recorded in the child
		PSP.  Note that break-checking may not be re-enabled until
		the int 24h address in the PSP has been restored.  */

	    if (_osmajor >= 5) {
		if (NOT bOrgUMBLink) {
		    _dos_setumblink (bOrgUMBLink);
		    _dos_setvect (0x24, lpOrgInt24h);
		    PSP_FP (childpsp) -> lpInt24h = lpOrgInt24h;
		    _dos_setbreakflag (bOrgBreakFlag);
		}
	    }

	    /*	By default, the child process is allowed access to memory
		which had been masked off during the loading stage - thus,
		the memory mask is released before the child's real code
		gets to run.  As a testing option however, the opportunity
		is given to restrict the child's execution to the loading
		area, as well.	*/

	    if (NOT bRestrictChildsMemory) ReleaseMask (&maskhandle);

	    if (bSetChildsStrategy) _dos_setstrategy (wChildsStrategy);
	    if (bSetChildsUMBLink) _dos_setumblink (bChildsUMBLink);

	    /*	If DOS is using the HMA, then the A20 line should be
		disabled when the child program starts executing.  DOS
		function 4B05h must be the last DOS function used before
		control is transferred.  Refer to the main text for a
		discussion on how this step has clearly not been planned
		carefully (if at all) in the design on DOS 5.  */

	    if (_osmajor >= 5) {
		execstate.wFlags = EXEC_OVERLAY;
		execstate.lpName = progname;
		execstate.spPSP = childpsp;
		execstate.lpEntryPoint = loadblock.lpEntryPoint;
		SetExecState ();
	    }

	    /*	Until now, a Ctrl-C or Ctrl-Break would have terminated
		the child process before any of its code (loaded from
		its file on disk) had got a chance to execute.	Mark
		the barrier with a Boolean variable so that bogus
		termination can be detected.  */

	    execution_attempted = TRUE;

	    _child = TRUE;

	    /*	Load the child program's stack.  A value to be passed in
		register AX must be popped off.  Place the child's entry
		point onto the stack, ready for transfer via a RETF
		instruction.  Make sure that DS and ES both contain the
		segment address of the child's PSP.  */

	    _asm {
		    mov     dx,childpsp
		    mov     ss, word ptr loadblock.lpStackTop+2
		    mov     sp, word ptr loadblock.lpStackTop
		    pop     ax
		    push    word ptr loadblock.lpEntryPoint+2
		    push    word ptr loadblock.lpEntryPoint
		    mov     es,dx
		    mov     ds,dx
		    retf
	    }		    // control resumes at the label termaddr (the
	    goto termaddr;  // (goto statement is merely a reminder), after
	}		    // which consultation of a Boolean variable
	else {		    // directs control to the following block

	    if (bSetChildsStrategy) _dos_setstrategy (wOrgStrategy);
	    if (bSetChildsUMBLink) _dos_setumblink (bOrgUMBLink);
	    if (bRestrictChildsMemory) ReleaseMask (&maskhandle);

	    /*	Retrieve the child's exit code:  after successful execution
		and termination of the child program, this is the value
		expected to be returned by the spawn functions.

		Beware however, of bogus termination caused by exceptions
		to DOS functions used between the initialisation of the
		child as a process (i.e., its loading) and the execution of
		the child's real program code.  In this case, there is no
		exit code from the child and the spawn functions should
		report that they were unable (for presumably temporary
		reasons) to start the child.  */

	    if (execution_attempted) {
		exitcode = GetChildsExitCode ();
	    }
	    else {
		errno = EAGAIN;
		exitcode = 0xFFFF;
	    }
	    convert_exitcode = FALSE;	    // these exit codes are final
	}
    }
    return (dosretax (exitcode, convert_exitcode));
}

void InitFCBs (char *cmdline)
{
    _asm {				// compiler preserves si, di
	    mov     si,cmdline
	    inc     si			// advance past length count
					// that starts command line
	    push    ds
	    pop     es
	    mov     di,offset fcb1
	    mov     ax,2901h
	    int     21h 		// DOS parse FCB function advances
					// si past argument (if present),
	    mov     di,offset fcb2	// ready for next
	    mov     ax,2901h
	    int     21h
    }
}

static _segment GetCurrentProcess (void)
{
    _asm {
	    mov     ah,51h
	    int     21h
	    mov     ax,bx
    }
}

static void SetExecState (void)
{
    _asm {
	    mov     dx,offset execstate
	    mov     ax,4B05h
	    int     21h
    }
}

static WORD GetChildsExitCode (void)
{
    _asm {
	    mov     ah,4Dh
	    int     21h
    }
}

/*  ************************************************************************  */

