
/*  ************************************************************************  *
 *				    a20map.c				      *
 *  ************************************************************************  */

#include    <dos.h>
#include    <stdarg.h>
#include    <stdio.h>
#include    <stdlib.h>

#include    <standard.h>
#include    <sysvars.h>

static CHAR description [] = "\n\
Determines the extended memory map when the A20 line is disabled.\n\
The present implementation has a granularity of one megabyte and\n\
requires an 80386 processor or better, operating in real mode.\n\
";

/*  Use XMS facilities for A20 manipulation.  */

extern void (_far *GetXMSAddress (void)) (void);

static void (_far *lpXMS) (void);

#define LocalEnableA20()    ((BOOL (_fastcall _far *) (WORD)) lpXMS) (0x0500)
#define LocalDisableA20()   ((BOOL (_fastcall _far *) (WORD)) lpXMS) (0x0600)
#define QueryA20()          ((BOOL (_fastcall _far *) (WORD)) lpXMS) (0x0700)

static WORD PhysicalDisableA20 (void);

/*  Use the real mode flat model to form addresses with more than 20 bits
    independently of any interface that enables the A20.  Unfortunately,
    this limits the program to use on an 80386 or better.  */

extern DWORD GetESLimit (void);
extern BOOL InProtectedMode (void);
extern BOOL Is80386 (void);
extern void SetESLimit (DWORD);

/*  Functions and variables for reading and writing test words at successive
    megabyte boudaries  */

static WORD GetTotalMemory (void);
extern WORD ReadID (WORD);
static CHAR *TestMemoryMap (WORD);
extern void WriteID (WORD, WORD);

#define	CAPACITY 16
static WORD disabled [CAPACITY], originals [CAPACITY];

/*  Message output  */

static void PutError (CHAR *, ...);

/*  ========================================================================  */

int main (int argc, char **argv)
{
    register CHAR *ptr;
    register WORD megabyte;
    WORD ceiling, totalmeminkb;
    DWORD limit;

    /*	Parse the command line.  */

    while (++ argv, -- argc) {
	ptr = *argv;
	if (*ptr == '/') {
	    if (ptr [1] == '?' AND ptr [2] == '\0') {
		printf (description);
		return (1);
	    }
	    PutError ("invalid switch - %s", ptr);
	    return (-1);
	}
	else {
	    PutError ("invalid parameter - %s", ptr);
	    return (-1);
	}
    }

    /*	All being well, check that the machine is an 80386 (so that the Flat
	Model may be used to access extended memory without relying on an
	interface that automatically enables the A20) and that an XMS server
	has been installed (for the means to control the A20 in the test).  */

    if (NOT Is80386 ()) {
	PutError ("requires 80386 CPU or better");
	return (-1);
    }
    if (InProtectedMode ()) {
	PutError ("CPU must be in real mode");
	return (-1);
    }
    lpXMS = GetXMSAddress ();
    if (lpXMS == NULL) {
	PutError ("no XMS server present");
	return (-1);
    }

    /*	Compute the number of 1MB boundaries that lie in extended memory
	known to be installed on the machine.  Although programs may trap
	int 15h function 88h to change the amount of extended memory
	reported to exist, the original value (disregarding possible
	activity before IO.SYS) is recorded by DOS - as a number of KB.  */

    totalmeminkb = GetTotalMemory ();
    if (totalmeminkb == 0) {
	PutError ("unable to determine amount of physical memory");
	return (-1);
    }
    ceiling = (totalmeminkb + 1023) >> 10;
    if (ceiling > CAPACITY) ceiling = CAPACITY;
    limit = ((DWORD) ceiling << 20) - 1;

    /*	Disable interrupts.  The particular scheme used for generating
	addresses above 1MB is unstable.  Moreover, the test involves
	writing a different value at each megabyte boundary for
	identification, saving whatever was there at the start - the small
	risk that an interrupting process may need this data should not be
	taken.	*/

    ptr = NULL; 	/*  The testing routine returns a (non-zero) pointer
			    to an error message - to be output once
			    interrupts are re-enabled.	*/

    _disable ();
    if (GetESLimit () < limit) SetESLimit (limit);

    /*	Enable the A20 line to begin the test and restore it afterwards.
	Note that the Local Enable function always activates the A20.  The
	matching Local Disable restores the previous state, which does not
	necessarily entail disabling the A20 physically.  */

    if (NOT LocalEnableA20 ()) ptr = "cannot enable A20";
    else {
	ptr = TestMemoryMap (ceiling);
	if (NOT LocalDisableA20 ())
	    if (ptr == NULL) ptr = "could not restore A20 after test";
    }

    _enable ();

    /*	Display the results - they cannot be presented at the time they are
	gathered because of the need to prevent interrupts.  Begin with any
	pending error message.  */

    if (ptr) {
	PutError (ptr);
	return (-1);
    }
    printf ("\nMemory mapping caused by disabling the A20 line:\n");
    for (megabyte = 0; megabyte < ceiling; megabyte ++) {
	printf ("\n  %2dMB -> %2dMB", megabyte, disabled [megabyte]);
    }
    printf ("\n");
    return (0);
}

static CHAR *TestMemoryMap (WORD ceiling)
{
    register WORD count, megabyte;

    /*	Write an identification number at each megabyte boundary, but first
	of all, save the data that will be overwritten.  This function is
	called with the A20 enabled, so the whole address space is
	accessible.  */

    for (megabyte = 0; megabyte < ceiling; megabyte ++) {
	originals [megabyte] = ReadID (megabyte);
    }
    for (megabyte = 0; megabyte < ceiling; megabyte ++) {
	WriteID (megabyte, megabyte);
    }

    /*	A single Local Disable function need not actually de-activate the
	A20 line.  This may be done however, by calling the Local Disable
	function repeatedly.  The number of calls needed should be counted
	so that the local A20 state can be restored later.  */

    count = PhysicalDisableA20 ();

    /*	If this has not in fact disabled the A20, then try to restore the
	XMS server's local A20 state before returning an error.  */

    if (NOT (QueryA20 () == OFF)) {
	if (count) {
	    do {
		if (LocalEnableA20 ()) count --;
	    } while (count);
	}
	return ("cannot disable A20");
    }

    /*	Now that the A20 line is known to be disabled, retrieve the words
	left at the megabyte boundaries.  */

    for (megabyte = 0; megabyte < ceiling; megabyte ++) {
	disabled [megabyte] = ReadID (megabyte);
    }

    /*	Enable the A20 again (restoring the local state) and replace the
	words that were originally at the megabyte boundaries.  */

    do {
	if (LocalEnableA20 ()) count --;
    } while (count);

    if (QueryA20 () == OFF) return ("cannot re-enable A20 to restore data");

    for (megabyte = 0; megabyte < ceiling; megabyte ++) {
	WriteID (megabyte, originals [megabyte]);
    }

    return (NULL);
}

static WORD PhysicalDisableA20 (void)
{
    register WORD count;

    /*	To disable the A20 using XMS functions, call the Local Disable A20
	service repeatedly.  This should eventually disable the true A20
	line, causing the next Local Disable to fail.

	Counting the Local Disable calls yields the XMS server's internal
	A20 counter - which is returned as this C function's value.  Each
	Local Disable must be matched by a Local Enable when restoring the
	A20 to its original state.

	Since the Local Disable call may fail because of an A20 hardware
	problem, the A20 status should be tested after this function
	returns - to confirm that the A20 has in fact been disabled.  */

    if (QueryA20 () == OFF) return (0);

    /*	Limit the number of times that LocalDisableA20 () is called - just
	in case it never fails.  */

    count = 0x0100;
    while (LocalDisableA20 () AND -- count) {
    }
    return (0x0100 - count);
}

static WORD GetTotalMemory (void)
{
    if (_osmajor < 4) return (0);

    _asm {
	    mov     ah,52h
	    int     21h
	    mov     ax,es:[bx].wExtendedMemoryInKB
	    add     ax,1024
	    jnc     done
	    xor     ax,ax
	done:
    }
}

static void PutError (CHAR *str, ...)
{
    va_list va_start (argptr, str);
    printf ("\nA20MAP error:  ");
    vprintf (str, argptr);
    printf ("\n");
    va_end (argptr);
}

/*  ************************************************************************  */

