
/*  ************************************************************************  *
 *		   sdconfig.c - primary file for sdconfig.exe		      *
 *  ************************************************************************  */

#include    <stdarg.h>
#include    <stdio.h>
#include    <stdlib.h>
#include    <string.h>

#include    <smartdrv.h>
#include    <standard.h>

static CHAR description [] = "\n\
Resizes and otherwise configures the SMARTDRV.SYS disk cache.\n\
\n\
  SDCONFIG [newsize|MIN|MAX] [OFF|ON|RESET] [/s]\n\
\n\
Without parameters, the program reports the current cache size and status.\n\
\n\
Otherwise, specify a new cache size in KB or a new status.  The /s switch\n\
directs the program to report the cache's condition after making changes.\n\
";

/*  SMARTDrive manipulations at DOS function level are handled externally -
    in SMARTDRV.C  */

extern int OpenSMARTDrive (void);
extern BOOL SupportsIOCTL (int);
extern BOOL ReadSMARTDrive (int, struct SD_READ *);
extern BOOL WriteSMARTDrive (int, struct SD_WRITE *);
extern BOOL CloseSMARTDrive (int);

void ShowStatus (struct SD_READ *);

static int sd_handle;
static struct SD_READ sd_read;
static struct SD_WRITE sd_write;

static void Cleanup (void);
static BOOL IsReallyZero (CHAR *);
static void PutError (CHAR *, ...);
static void PutNewLine (void);

int main (int argc, char **argv)
{
    register CHAR *ptr;
    register WORD newsize;
    BOOL max = FALSE, min = FALSE, sizeoption = FALSE;
    BOOL off = FALSE, on = FALSE, reset = FALSE, statusoption = FALSE;
    BOOL showstatus = FALSE;

    while (++ argv, -- argc) {
	ptr = *argv;
	if (*ptr == '/') {
	    ptr ++;

	    /*	A description of the program may be obtained by supplying
		the /? switch as usual.  */

	    if (ptr [0] == '?' AND ptr [1] == '\0') {
		printf (description);
		return (0);
	    }

	    /*	The /s switch indicates that the cache status should be
		shown after writing to the device - to allow verification
		of any changes.  */

	    if ((ptr [0] | 0x20) == 's' AND ptr [1] == '\0') {
		showstatus = TRUE;
		continue;
	    }

	    /*	No other switches are accepted.  */

	    ptr --;
	    PutError ("invalid switch - %s", ptr);
	    return (-1);
	}
	else {
	    if (NOT statusoption) {
		statusoption = TRUE;
		if (stricmp ("OFF", ptr) == 0) {
		    off = TRUE;
		    continue;
		}
		if (stricmp ("ON", ptr) == 0) {
		    on = TRUE;
		    continue;
		}
		if (stricmp ("RESET", ptr) == 0) {
		    reset = TRUE;
		    continue;
		}
		statusoption = FALSE;
	    }

	    /*	A new cache size is expected as a single numerical
		parameter.  */

	    if (NOT sizeoption) {
		sizeoption = TRUE;

		newsize = atoi (ptr);
		if (newsize != 0 OR IsReallyZero (ptr)) {

		    /*	Scale by 16 since SMARTDrive measures its size in
			multiples of 16KB.  */

		    newsize >>= 4;
		    continue;
		}

		/*  For convenience, allow MIN and MAX keywords.  */

		if (stricmp ("MIN", ptr) == 0) {
		    min = TRUE;
		    continue;
		}
		if (stricmp ("MAX", ptr) == 0) {
		    max = TRUE;
		    continue;
		}
		sizeoption = FALSE;
	    }

	    /*	No other parameters are accepted.  */

	    PutError ("invalid parameter - %s", ptr);
	    return (-1);
	}
    }

    /*	Open the SMARTAAR device.  */

    sd_handle = OpenSMARTDrive ();
    if (sd_handle == -1) {
	PutError ("cannot open SMARTAAR device");
	return (-1);
    }

    /*	DOS will of course, close the now open device when the program
	terminates, so there is no real need to close the device explicitly
	at the end of every decision path taken by the program from here on.
	On the other hand, it is undeniably beneficial to at least be aware
	that, in general, programs must allow for cleaning up activities
	once they obtain a system resource.  Strictly speaking, programs
	should trap the DOS exceptions, as well.  */

    atexit (Cleanup);

    /*	Check that the handle does actually refer to a character device
	supporting an IOCTL interface, then read the data packet that
	describes SMARTDrive's configuration.  */

    if (NOT SupportsIOCTL (sd_handle)) {
	PutError ("name conflict with file or unrecognised SMARTAAR device");
	return (-1);
    }

    if (NOT ReadSMARTDrive (sd_handle, &sd_read)) {
	PutError ("cannot control SMARTAAR device");
	return (-1);
    }

    /*	In the absence of command line options, generate a status report and
	return.  */

    if (NOT (sizeoption OR statusoption)) {
	ShowStatus (&sd_read);
	return (0);
    }

    if (sizeoption) {

	/*  Now that the cache configuration is known, resolve MIN or MAX or
	    check that the numerical value given for the new cache size is
	    in fact within the allowed range.  */

	if (min) newsize = sd_read.MinimumSize;
	else if (max) newsize = sd_read.MaximumSize;
	else {
	    if (newsize < sd_read.MinimumSize
		    OR newsize > sd_read.MaximumSize) {
		PutError ("desired size %uKB is out of range (%u to %u)",
			newsize << 4,
			sd_read.MinimumSize << 4, sd_read.MaximumSize << 4);
		return (-1);
	    }
	}

	/*  Set the parameters for the write packet.  Skip the step of
	    writing to the device if the cache size is not to change.  */

	if (newsize != sd_read.CurrentSize) {
	    if (newsize < sd_read.CurrentSize) {
		sd_write.data.size = sd_read.CurrentSize - newsize;
		sd_write.command = SD_DECREASE_CACHE;
	    }
	    else {
		sd_write.data.size = newsize - sd_read.CurrentSize;
		sd_write.command = SD_INCREASE_CACHE;
	    }

	    /*	Write the new control information to the SMARTAAR device.  */

	    if (NOT WriteSMARTDrive (sd_handle, &sd_write)) {
		PutError ("unable to change cache size");
		return (-1);
	    }
	}

	/*  If the /s switch was given, read the configuration back from
	    the device for an up-to-date status report.  */

	if (showstatus) {
	    if (NOT ReadSMARTDrive (sd_handle, &sd_read)) showstatus = FALSE;
	}
    }

    if (statusoption) {
	if (off) sd_write.command = SD_DEACTIVATE;
	else if (on) sd_write.command = SD_ACTIVATE;
	else if (reset) sd_write.command = SD_RESET;

	if (NOT WriteSMARTDrive (sd_handle, &sd_write)) {
	    PutError ("unable to change cache status");
	    return (-1);
	}
    }

    /*	If the /s switch was given, read the configuration back from the
	device for an up-to-date status report.  */

    if (showstatus AND ReadSMARTDrive (sd_handle, &sd_read))
	ShowStatus (&sd_read);

    return (0);
}

void ShowStatus (register struct SD_READ *sdrptr)
{
    printf ("\nSMARTDrive version %u.%02u installed %s:",
	    sdrptr -> MajorVersion, sdrptr -> MinorVersion,
	    sdrptr -> IsActive ? "and active" : "but deactivated");
    printf ("\nCurrently using %uKB of %s memory",
	    sdrptr -> CurrentSize << 4,
	    sdrptr -> MemoryType == 1 ? "extended" : "expanded");

    PutNewLine ();

    printf ("\nMinimum cache size is %5uKB", sdrptr -> MinimumSize << 4);
    printf ("\nMaximum cache size is %5uKB", sdrptr -> MaximumSize << 4);

    PutNewLine ();

    printf ("\nCache has room for %u tracks", sdrptr -> TracksInCache);
    printf ("\nand presently contains %u valid tracks",
	    sdrptr -> CurrentTracks);

    PutNewLine ();

    printf ("\nCache performance this session:  ");
    printf ("\n  %2u%% of sector reads filled from main cache",
	    sdrptr -> HitRatio);
    printf ("\n  %2u%% of sector reads filled from intermediate buffer",
	    sdrptr -> BufferRatio);

    PutNewLine ();
}

static void Cleanup (void)
{
    CloseSMARTDrive (sd_handle);
}

static BOOL IsReallyZero (register CHAR *ptr)
{
    CHAR ch;

    do {
	ch = *ptr ++;
    } while (ch == '0');
    return (ch == '\0' ? TRUE : FALSE);
}

static void PutNewLine (void)
{
    printf ("\n");
}

void PutError (CHAR *str, ...)
{
    va_list va_start (argptr, str);
    printf ("\nSMARTDRV fatal error:  ");
    vprintf (str, argptr);
    PutNewLine ();
    va_end (argptr);
}

/*  ************************************************************************  */

