
/*  ************************************************************************  *
 *				   smartdrv.c				      *
 *  ************************************************************************  */

/*  This file (which may be compiled as a resident module under the Keep C
    scheme for resident programming) contains routines for manipulating the
    SMARTDrive disk cache, both for small steps (open, read, write, close)
    and for the particular task of resizing the cache in extended memory.

    To adapt the routines for use when control has already passed through
    the DOS kernel, i.e., to demonstrate the undocumented callback facility
    provided by int 2Fh function 12h, define the symbol _CALLBACK_ at
    compile time and use the segmentation model for resident programming.  */

#include    <driver.h>
#include    <keepc.h>
#include    <sft.h>
#include    <smartdrv.h>
#include    <standard.h>

/*  The logical device name by which SMARTDrive is known to DOS.  */

static CHAR smartaar [] = "SMARTAAR";

/*  Transfer buffers for SMARTDrive control data.  */

static struct SD_READ sd_read = {0};
static struct SD_WRITE sd_write = {0};

/*  For the convenience of external modules, separate functions are provided
    to make the SMARTDrive cache grow or shrink (by some number of KB).  */

BOOL DecreaseCache (DWORD);
BOOL IncreaseCache (DWORD);

/*  Both those functions behave similarly in terms of their access to
    SMARTDrive, so they are passed to a single function that carries a
    variable (the SMARTDrive command number) to distinguish expansion from
    contraction.  */

static BOOL ChangeCache (BYTE, DWORD);

/*  Separate C functions for each call to the SMARTAAR device.	These are
    written with in-line assembler to demonstrate the correspondence between
    the int 21h functions and their callback equivalents.  Registers may be
    corrupted during DOS callback, but as written here, these C functions
    preserve only ds and bp explicitly - the liability of si and di to
    change is fairly easily accommodated by the caller, as shown by the
    design of ChangeCache ().  */

int OpenSMARTDrive (void);
BOOL SupportsIOCTL (int);
BOOL ReadSMARTDrive (int, struct SD_READ *);
BOOL WriteSMARTDrive (int, struct SD_WRITE *);
BOOL CloseSMARTDrive (int);

#ifdef	_CALLBACK_

/*  Special preparation required for the callback interface.  */

static BOOL IsCallbackOK (void);

/*  For strict safety, some material in the DOS kernel's data area must be
    preserved across the call to SMARTDrive.  Otherwise, when the kernel
    opens the SMARTAAR device, it will lose the address of the SFT it has
    opened for the XMSFILEX device.  */

static void SaveKernelData (void);
static void RestoreKernelData (void);
struct SFT _far * _far *lpSFTPointer = NULL;

extern struct REQUEST_HEADER _far *lpRequestHeader;	// in driver.asm

#endif

/*  ========================================================================  */

BOOL IncreaseCache (DWORD increment)
{
    return (ChangeCache (SD_INCREASE_CACHE, increment));
}

BOOL DecreaseCache (DWORD decrement)
{
    return (ChangeCache (SD_DECREASE_CACHE, decrement));
}

/*  ------------------------------------------------------------------------  */

static BOOL ChangeCache (BYTE command, DWORD change)
{
    register WORD change_16KB, latitude;	// register keyword ensures
    int handle; 				// preservation of si and di
    BOOL cache_changed = FALSE;

    /*	The SMARTDrive cache size is measured and manipulated in multiples
	of 16KB, using a 16-bit variable, so check that the desired change
	(supplied in KB) is within range.  */

    if (HIGHWORD (change) >> 4) return (FALSE);

    /*	The callback facility assumes an internal DOS stack - check that
	the current stack belongs to the DOS kernel.  */

    #ifdef  _CALLBACK_
    if (NOT IsCallbackOK ()) return (FALSE);
    SaveKernelData ();
    #endif

    /*	Open the SMARTAAR device.  */

    handle = OpenSMARTDrive ();
    if (handle == -1) return (FALSE);

    /*	Check that the resulting handle does in fact refer to a character
	device that supports IOCTL, then read SMARTDrive's configuration.  */

    if (SupportsIOCTL (handle) AND ReadSMARTDrive (handle, &sd_read)) {

	/*  If the cache is in extended memory, calculate the increment
	    or decrement in multiples of 16KB and verify that SMARTDrive
	    is able to accommodate the change.	*/

	if (sd_read.MemoryType == 0x01) {

	    change_16KB = (HIGHWORD (change) << 12) + (LOWWORD (change) >> 4);

	    if (change_16KB != 0) {

		latitude = command == SD_DECREASE_CACHE ?
			sd_read.CurrentSize - sd_read.MinimumSize :
			sd_read.MaximumSize - sd_read.CurrentSize;

		if (change_16KB <= latitude) {

		    /*	Send the desired cache change to SMARTDrive via the
			IOCTL Write interface.	*/

		    sd_write.command = command;
		    sd_write.data.size = change_16KB;

		    if (WriteSMARTDrive (handle, &sd_write))
			cache_changed = TRUE;
		}
	    }
	}
    }
    CloseSMARTDrive (handle);

    #ifdef  _CALLBACK_
    RestoreKernelData ();
    #endif

    return (cache_changed);
}

/*  ------------------------------------------------------------------------  */

#ifdef	_CALLBACK_

static BOOL IsCallbackOK (void)
{
    _asm {

    /*	Return TRUE if and only if the current stack belongs to the DOS
	kernel.  The segment address of the kernel's data and stack is
	returned in register ds by int 2Fh function 1203h (one of the
	few callback services that do not assume an internal DOS stack).  */

	    push    ds
	    push    bp
	    mov     ax,1203h
	    int     2Fh
	    mov     ax,ds
	    mov     dx,ss
	    pop     bp
	    pop     ds
	    cmp     ax,dx
	    jnz     bad

	    mov     ax,0001h	; TRUE
	    jmp     done
	bad:
	    xor     ax,ax	; FALSE
	done:
    }
}

#endif

int OpenSMARTDrive (void)
{
    _asm {

    /*	Open the SMARTDrive character device (SMARTAAR) for read/write
	access, returning the handle if successful, or else FFFFh.  Callback
	function 26h invokes the same code used by DOS function 3Dh to open
	a file.  */

	    mov     dx,OFFSET smartaar

    #ifndef _CALLBACK_

	    mov     ax,3D02h
	    int     21h
    #else
	    mov     cl,02h	; access flags in cl instead of al
	    push    ds
	    push    bp
	    mov     ax,1226h
	    int     2Fh
	    pop     bp
	    pop     ds
    #endif
	    jnc     done
	    mov     ax,0FFFFh
	done:
    }
}

BOOL SupportsIOCTL (int handle)
{
    _asm {

    /*	Return TRUE if and only if the handle refers to a character device
	that responds to IOCTL.  The whole int 21h function 44h family is
	supported by callback function 2Bh - the subfunction number usually
	in AL is passed in the low byte of BP instead.  */

	    mov     bx,handle

    #ifndef _CALLBACK_

	    mov     ax,4400h	; subfunction 00h - Get Device Status
	    int     21h
    #else
	    push    ds
	    push    bp
	    mov     bp,4400h
	    mov     ax,122Bh
	    int     2Fh
	    pop     bp
	    pop     ds
    #endif
	    jc	    bad
	    and     dx,4080h	; 0080h is mask for character device,
	    cmp     dx,4080h	; 4000h for IOCTL SUpport
	    jnz     bad
	    mov     ax,0001h	; TRUE
	    jmp     done
	bad:
	    xor     ax,ax	; FALSE
	done:
    }
}

BOOL ReadSMARTDrive (int handle, struct SD_READ *sd)
{
    _asm {

    /*	Return TRUE if and only if an IOCTL Read from SMARTDrive (of the
	whole configuration structure) succeeds.  */

	    mov     dx,sd
	    mov     cx,size sd_read
	    mov     bx,handle

    #ifndef _CALLBACK_

	    mov     ax,4402h	; subfunction 02h - IOCTL Read from device
	    int     21h
    #else
	    push    ds
	    push    bp
	    mov     bp,4402h
	    mov     ax,122Bh
	    int     2Fh
	    pop     bp
	    pop     ds
    #endif
	    jc	    bad
	    cmp     ax,cx
	    jnz     bad
	    mov     ax,0001h	; TRUE
	    jmp     done
	bad:
	    xor     ax,ax	; FALSE
	done:
    }
}

BOOL WriteSMARTDrive (int handle, struct SD_WRITE *sd)
{
    _asm {
	    mov     dx,sd
	    mov     cx,size sd_write
	    mov     bx,handle

    #ifndef _CALLBACK_

	    mov     ax,4403h	; subfunction 03h - IOCTL Write to device
	    int     21h
    #else
	    push    ds
	    push    bp
	    mov     bp,4403h
	    mov     ax,122Bh
	    int     2Fh
	    pop     bp
	    pop     ds
    #endif
	    jc	    bad
	    cmp     ax,cx
	    jnz     bad
	    mov     ax,0001h	; TRUE
	    jmp     done
	bad:
	    xor     ax,ax	; FALSE
	done:
    }
}

BOOL CloseSMARTDrive (int handle)
{
    _asm {

    /*	Close the file handle, using DOS function 3Eh as usual or callback
	function 27h.  */

	    mov     bx,handle

    #ifndef _CALLBACK_

	    mov     ah,3Eh
	    int     21h
    #else
	    push    ds
	    push    bp
	    mov     ax,1227h
	    int     2Fh
	    pop     bp
	    pop     ds
    #endif
	    jc	    bad
	    mov     ax,0001h	; TRUE
	    jmp     done
	bad:
	    xor     ax,ax	; FALSE
	done:
    }
}

/*  ------------------------------------------------------------------------  */

#ifdef	    _CALLBACK_

/*  Temporary storage for the kernel's pointer to the SFT opened for the
    XMSFILEX device  */

static struct SFT _far *lpXMSFILEX_SFT = 0;

/*  The REQUEST_HEADER structure contains un-named substructures and cannot
    easily be declared as initialised data, hence the following ugly
    construction for a temporary area in which to save the device driver
    header passed to the XMSFILEX device while asking DOS to call
    SMARTDrive.  */

static BYTE dummy [sizeof (struct REQUEST_HEADER)] = {0};
#define XMSFILEX_ReqHdr *((struct REQUEST_HEADER *) dummy)

void SaveKernelData (void)
{
    lpXMSFILEX_SFT = *lpSFTPointer;
    XMSFILEX_ReqHdr = *lpRequestHeader;
}

void RestoreKernelData (void)
{
    *lpSFTPointer = lpXMSFILEX_SFT;
    *lpRequestHeader = XMSFILEX_ReqHdr;
}

#endif

/*  ************************************************************************  */

