
/*  ************************************************************************  *
 *		    main.c - initialisation file for XMSFILE		      *
 *  ************************************************************************  */

#include    <dos.h>
#include    <stdarg.h>
#include    <stdio.h>
#include    <stdlib.h>
#include    <string.h>

#include    <driver.h>
#include    <keepc.h>
#include    <standard.h>

static CHAR description [] = "\n\
XMSFILE is a character device driver that uses XMS to emulate a file.\n\
\n\
    device = xmsfile.exe [clustersize] [/Local] [/SMARTDrive]\n\
\n\
Once installed, the device acts as a clipboard - with the name XMSFILEX.\n\
\n\
The allocation unit (clustersize) is measured in kilobytes\n\
and must be a power of 2.  The default is 64KB.\n\
\n\
Attempts may be made to reclaim extended memory from SMARTDrive\n\
if the cluster size is at least 16 and the /s switch is given.\n\
\n\
The device is global by default, that is, an XMS handle is obtained during\n\
initialisation and although memory is surrendered when the device is not in\n\
use, the handle is always kept.  The /l switch configures the device to\n\
seek an XMS handle only when one is needed and to release it when the\n\
clipboard is emptied.  In this configuration, the device behaves properly\n\
under Windows enhanced mode if and only if a Local=XMSFILEX string is\n\
placed in the [386Enh] section of SYSTEM.INI and the device is not already\n\
in use when Windows is started;  each virtual machine accesses its own\n\
clipboard, hence its description as local.\n";

/*  A flag defined by the CRTDRVR library to enable programs to determine
    whether or not they are being run to initialise a device driver or as
    a normal DOS executable  */

extern BOOL _RESIDENT bLoadedAsDriver;

/*  A function to find the address of the XMS driver's calling point and a
    resident variable in which to record this address  */

extern void (_far *GetXMSAddress (void)) (void);
extern void (_far * _RESIDENT lpXMS) (void);

/*  A parameter describing the XMSFILE's allocation unit  */

extern WORD _RESIDENT wLogClusterSize;

/*  A resident function to obtain an initial memory allocation by opening
    the device.  */

extern WORD _far FarOpen (void);

/*  A flag denoting that extended memory should be recovered from SMARTDrive
    should no memory be available by normal means - and a function to
    prepare this facility  */

extern BOOL _RESIDENT bUseSMARTDrive;
extern BOOL InitSMARTDrive (WORD);

/*  A flag in resident memory to record that the driver should attempt to
    behave locally to each Windows virtual machine - plus a warning message  */

extern BOOL _RESIDENT bLocal;

static char localwarning [] = "\n\
For correct behaviour under Windows enhanced mode, the device requires\n\
the string Local=XMSFILEX in the [386Enh] section of SYSTEM.INI\n\
and must not be in use when Windows is started\n";

/*  Miscellany  */

static WORD log (WORD);
void PutError (char *, ...);
void PutWarning (char *, ...);

/*  ========================================================================  */

int main (int argc, char **argv)
{
    register CHAR *ptr;
    register WORD clustersizeinkb;
    BOOL clustersizedone = FALSE, smartdrivedone = FALSE;
    WORD temp;
    struct find_t fileinfo;

    /*	If XMSFILE.EXE is executed as a DOS program, rather than as a device
	driver, display a message describing its function and syntax.  */

    if (NOT bLoadedAsDriver) {
	printf (description);
	return (0);
    }

    /*	Avoid multiple loading:  abort the installation if an XMSFILEX
	device or file exists already.	*/

    #define _A_GENERAL	(_A_HIDDEN | _A_SYSTEM | _A_SUBDIR)

    if (_dos_findfirst ("XMSFILEX", _A_GENERAL, &fileinfo) == 0) {
	PutError ("XMSFILEX already present");
	return (-1);
    }

    /*	Parse the command line.  */

    while (++ argv, -- argc) {
	ptr = *argv;
	if (*ptr == '/') {
	    ptr ++;

	    /*	The switches understood by the program need not be
		entered in full - any initial substring is acceptable.  */

	    if (strnicmp ("SMARTDrive", ptr, strlen (ptr)) == 0) {
		smartdrivedone = TRUE;
		continue;
	    }
	    if (strnicmp ("Local", ptr, strlen (ptr)) == 0) {
		bLocal = TRUE;
		continue;
	    }
	    ptr --;
	    PutError ("invalid switch %s", ptr);
	    return (-1);
	}
	else {
	    if (clustersizedone) {
		PutError ("too many parameters");
		return (-1);
	    }

	    /*  A numerical parameter must be supplied - XMSFILE's memory
		allocation will be adjusted in multiples of this many KB,
		defaulting to 64.  */

	    clustersizeinkb = atoi (ptr);
	    if (clustersizeinkb == 0) {
		PutError ("expected cluster size, got %s", ptr);
		return (-1);
	    }
	    temp = log (clustersizeinkb);
	    if (temp == -1) {
		PutError ("cluster size must be power of 2");
		return (-1);
	    }
	    wLogClusterSize = temp + 10;
	    clustersizedone = TRUE;
	}
    }
    if (NOT clustersizedone) {	    // the default cluster size must be 64KB
	clustersizeinkb = 64;	    // to ensure proper behaviour of XMS
	wLogClusterSize = 16;	    // function 0Fh in versions 2.xx of
    }				    // HIMEM.SYS.

    /*	If the option to steal memory from SMARTDrive (when necessary) has
	been exercised, check its feasibility.	It is the responsibility of
	the SMARTDrive-specific testing function to issue any warning
	message and disable the option, if conditions are not suitable.  */

    if (smartdrivedone) bUseSMARTDrive = InitSMARTDrive (clustersizeinkb);

    /*	Check that an XMS server is installed and if so, record its calling
	address.  */

    lpXMS = GetXMSAddress ();
    if (lpXMS == NULL) {
	PutError ("no XMS server installed");
	return (-1);
    }

    /*	Allocate an Extended Memory Block.  This step would not be needed in
	a single-tasking environment or if task-switching were fully
	supported (since in both cases, the driver could simply wait for an
	attempt to open the device).  */

    if (NOT bLocal) {
	if (FarOpen () & D_ERROR) {
	    PutError ("Extended memory handle unavailable");
	    return (-1);
	}
    }

    /*	Indicate successful installation, then tidy up.  */

    printf ("\nXMSFILEX installed.  Cluster size = %dKB.\n", clustersizeinkb);
    if (bLocal) printf (localwarning);

    CloseAllFiles ();
    FreeEnvironment ();
    ExitKeepC (0, 0);
}

/*  The logarithm of a power of 2 may be computed easily enough by shifting
    bits to count the number of trailing zeros.  The function returns the
    impossible value 0xFFFF if given an argument that is not a power of 2.  */

static WORD log (register WORD x)
{
    register WORD n = 0;

    if (x) {
	while ((x & 1) == 0) {
	    x >>= 1;
	    n ++;
	}
	if (x == 1) return (n);
    }
    return (0xFFFF);
}

void PutError (char *str, ...)
{
    va_list va_start (argptr, str);
    printf ("\nXMSFILE intialisation error:  ");
    vprintf (str, argptr);
    printf ("\n");
    va_end (argptr);
}

void PutWarning (char *str, ...)
{
    va_list va_start (argptr, str);
    printf ("\nXMSFILE warning:  ");
    vprintf (str, argptr);
    printf ("\n");
    va_end (argptr);
}

/*  ************************************************************************  */

