
/*  ************************************************************************  *
 *				  transfer.c				      *
 *  ************************************************************************  */

/*  This file (suitable for use as a resident module under the Keep C scheme
    for resident programming) accompanies the character device driver
    emulation of a sequential file in extended memory.	It contains the
    material that depends use of the XMS to manage extended memory blocks.  */

#include    <driver.h>
#include    <keepc.h>
#include    <standard.h>

#include    "xmsfunc.h"

/*  Variables describing the file state - declared in SEQFILE.C  */

extern DWORD dwFilePtr;
extern DWORD dwFileEnd;

/*  The XMS calling address - recorded at initialisation.  */

void (_far *lpXMS) (void) = NULL;

/*  Variables describing the Extended Memory Block (EMB) obtained from the
    XMS server	*/

static DWORD dwEMBSizeInBytes = 0;
static DWORD dwEMBSizeInKB = 0;
static WORD wEMBHandle = 0;

/*  A structure for the XMS function that moves data to and from extended
    memory  */

static struct _XMS_MOVEBLOCK {
    DWORD length;
    WORD source_handle;
    DWORD source_offset;
    WORD dest_handle;
    DWORD dest_offset;
} moveblock = {0};

/*  A parameter describing the allocation unit as a binary logarithm of the
    size in KB - this is recorded at initialisation.  */

WORD wLogClusterSize = 16;		//  default cluster size is 64KB
				
/*  A Boolean variable which is FALSE by default but may be set to TRUE
    (during initialisation) if extended memory should be sought from the
    SMARTDrive disk cache when desperate.  */

BOOL bUseSMARTDrive = FALSE;

/*  The total amount (measured in KB) by which the cache has been reduced -
    maintained so that an attempt can be made to restore the cache when the
    device is closed.  */

static WORD wCacheDeficit = 0;

/*  External functions for manipulating the SMARTDrive cache  */

extern BOOL DecreaseCache (DWORD);
extern BOOL IncreaseCache (DWORD);

/*  A flag recorded during initialisation which determines whether the XMS
    handle should be released whenever the device is not in active use	*/

BOOL bLocal = FALSE;

/*  ========================================================================  */

/*  SEQFILE.C requires the following function in order to know how to
    transfer data.  The command argument may be D_READ or D_WRITE to
    indicate whether data is being read from or written to extended
    memory respectively.  */

BOOL MoveBlock (BYTE command, BYTE _far *lpBuffer, DWORD xmptr, WORD count)
{
    /*	The XMS interface uses a natural description of the transfer
	regions, so the function need simply record its arguments in the
	parameter block.  The EMB is accessed through a handle - the linear
	address of the block need not be known.  The buffer in real-mode
	addressable memory is described by a null handle.  */

    moveblock.length = count;
    if (command == D_READ) {
	moveblock.source_handle = wEMBHandle;
	moveblock.source_offset = xmptr;
	moveblock.dest_handle = 0;
	moveblock.dest_offset = (DWORD) lpBuffer;
    }
    else if (command == D_WRITE) {
	moveblock.source_handle = 0;
	moveblock.source_offset = (DWORD) lpBuffer;
	moveblock.dest_handle = wEMBHandle;
	moveblock.dest_offset = xmptr;
    }
    else return (FALSE);

    /*	Call the XMS server, returning a Boolean status (ignoring any XMS
	error code).  */

    _asm {
	    mov     si,OFFSET moveblock     ; the compiler preserves si
	    mov     ah,0Bh
	    call    lpXMS
    }
}

/*  The function AdjustRequest is called by SEQFILE to determine how much of
    a read or write request can actually be met.  Its return value is
    necessarily less than or equal to the DWORD argument.  If asked to write
    more data than the EMB can accommodate, this XMS version of the function
    will try to expand the current allocation.

    Note however that this implementation is fairly simplistic and expects
    that the caller will be interested mainly in whether or not the whole
    request can be accommodated.  Thus, if meeting the request demands
    expansion of the EMB, which turns out to be impossible, the present
    implementation does not try expanding the EMB gradually but simply trims
    the request to the amount of unused space currently in the EMB.  Against
    this, the function tries very hard to arrange the EMB's expansion, even
    to the extent of stealing memory from the SMARTDRV.SYS disk cache.	*/

DWORD AdjustRequest (BYTE command, DWORD numwanted)
{
    DWORD avail, extrabytes, extraclusters, extrakb, newsizeinkb;
    WORD cache_tried;

    /*	If the driver has no EMB to read from or write to, then stop the
	request.  */

    if (wEMBHandle == 0) return (0);

    /*	Compare the request against the amount of available space.  Read
	operations are limited to the area of the file to which data has
	already been written.  */

    avail = (command == D_READ ? dwFileEnd : dwEMBSizeInBytes) - dwFilePtr;
    if (numwanted > avail) {

	/*  Simply reduce the size of any request to read past the end of
	    the file.  */

	if (command == D_READ) return (numwanted = avail);

	/*  The XMS offers the opportunity of resizing the EMB to
	    accommodate an attempt to write more data than will fit.
	    Control the file's growth by allocating extra memory in
	    multiples of some cluster size fixed during initialisation (to
	    be some power of two when measured in kilobytes).  A larger
	    cluster size will mean fewer calls to the XMS server but may
	    also mean using more extended memory than necessary.  */

	extraclusters = shr (numwanted - avail - 1, wLogClusterSize) + 1;
	extrabytes = shl (extraclusters, wLogClusterSize);
	extrakb = shr (extrabytes, 10);

	/*  Only with version 3.0 of the XMS interface is it possible to
	    accommodate block sizes of 64MB or more.  This demonstration
	    does not use the 32-bit XMS function 8Fh introduced to deal
	    with such large EMBs - partly because the arrangements required
	    under Microsoft C 6.0 to use 32-bit registers for the XMS call
	    are prohibitive, but mostly because the extant implementation
	    (HIMEM.SYS version 3.07 distributed with Windows 3.1) is
	    seriously flawed.  */

	newsizeinkb = dwEMBSizeInKB + extrakb;
	if (HIGHWORD (newsizeinkb) != 0) return (numwanted = avail);

	/*  If the first attempt to resize the EMB fails, it may be
	    worthwhile seeking the extra space from the SMARTDrive disk
	    cache and then to retry.  */

	cache_tried = 0;
	for (;;) {
	    if (ResizeEMB (wEMBHandle, (WORD) newsizeinkb)) {
		dwEMBSizeInBytes += extrabytes;
		dwEMBSizeInKB += extrakb;
		break;
	    }
	    if (bUseSMARTDrive) {
		cache_tried ++;
		if (cache_tried == 1) {

		    /*	It is preferable to ask the cache to surrender only
			the marginal requirement, but this will probably
			only work if the disk cache is directly below the
			XMSFILE EMB (or trivially, if XMSFILE has no EMB).  */

		    if (DecreaseCache (extrakb)) {
			wCacheDeficit += extrakb;
			continue;
		    }
		}
		else if (cache_tried == 2) {

		    /*	If it turns out that the block still cannot be
			expanded, then try shrinking the cache so far that
			the memory released will be enough to contain an
			enlarged XMSFILE block.  */

		    if (DecreaseCache (dwEMBSizeInKB)) {
			wCacheDeficit += dwEMBSizeInKB;
			continue;
		    }
		}
	    }

	    /*	If all this cache manipulation fails to help, tailor the
		request to fit the available space.  */

	    numwanted = avail;
	    break;
	}
    }
    return (numwanted);
}

/*  Device driver functions supplied in SEQFILE.C for the open and close
    commands ensure that the file pointer is reset when the device is opened
    and that the file is truncated at the last position read or written when
    the device is closed.  Open/close features specific to the method of
    extended memory management are provided by the following functions.  */

WORD Open (void)
{
    DWORD temp;

    if (wEMBHandle == 0) {
	dwFileEnd = 0;
	dwEMBSizeInBytes = 0;
	dwEMBSizeInKB = 0;
	temp = AllocateEMB (0);
	if (NOT (BOOL) temp) return (D_ERROR | D_DONE | D_GENERAL_FAILURE);
	wEMBHandle = HIGHWORD (temp);
    }
    return (D_DONE);
}

WORD _far FarOpen (void)		// provided so that the device may
{					// be opened during initialisation
    return (Open ());			// if configured to act globally
}

WORD Close (void)
{
    DWORD newclusters, newkb, newsize, returntocache;

    if (wEMBHandle == 0) return (D_DONE | D_ERROR | D_GENERAL_FAILURE);

    /*	On closure, trim the EMB of any extended memory that is not being
	used to hold data.  */

    newclusters = dwFileEnd ? shr (dwFileEnd - 1, wLogClusterSize) + 1 : 0;
    newsize = newclusters ? shl (newclusters, wLogClusterSize) : 0;
    newkb = newsize ? shr (newsize, 10) : 0;

    if (newkb < dwEMBSizeInKB OR (newkb == 0 AND dwEMBSizeInKB == 0)) {

	returntocache = 0;

	/*  For a local device, the EMB handle can be released altogether
	    (and the wEMBHandle variable cleared) whenever the size returns
	    to zero.  */

	if (bLocal AND newkb == 0 AND ReleaseEMB (wEMBHandle)) {
	    wEMBHandle = 0;
	    returntocache = dwEMBSizeInKB;
	}
	else if (ResizeEMB (wEMBHandle, (WORD) newkb)) {
	    returntocache = dwEMBSizeInKB - newkb;
	    dwEMBSizeInKB = newkb;
	    dwEMBSizeInBytes = newsize;
	}
	else return (D_DONE | D_ERROR | D_GENERAL_FAILURE);

	/*  If the SMARTDrive cache has been shrunk to obtain more memory
	    for the file emulation, then attempt to return as much of this
	    as possible out of the amount just released.  */

	if (wCacheDeficit AND returntocache) {
	    if (returntocache < wCacheDeficit) returntocache = wCacheDeficit;
	    if (IncreaseCache (returntocache)) wCacheDeficit -= returntocache;
	}
    }
    return (D_DONE);
}

WORD Destroy (void)
{
    /*	This function is needed only for an intelligent local device.  It
	would be used on notification from the switcher that a task is being
	destroyed.  */

    Close ();		    //	destruction may pre-empt explicit closure
    _asm {
	    xor     dx,dx
	    xchg    dx,wEMBHandle
	    mov     ah,0Ah
	    call    lpXMS
    }
    return (D_DONE);
}

/*  ************************************************************************  */

