
/*  ************************************************************************  *
 *				   watched.c				      *
 *  ************************************************************************  */

#include    <stdio.h>

#include    <standard.h>

#include    "xmsfunc.h"
#include    "xmswatch.h"

/*  XMS services  */

extern void (_far *lpXMS) (void);

extern BOOL LockEMB (WORD, DWORD *);
extern BOOL GetEMBInfo (WORD, struct EMBINFO *);

/*  Two variables and one function for retrieving information from a
    resident copy of XMSWATCH  */

static struct LIST *pWatchedEMBs;
static struct XMSWATCH xmswatch;

extern struct LIST *ConsultWatcher (struct XMSWATCH *);

/*  Results are presented in a multi-column table under the following
    header.  Data for each column is handled by a separate function that
    pads the item to the expected width - adjacent columns are separated
    by five spaces.  */

static CHAR banner [] = "\n\
Extended Memory Allocation Detected by Resident Component of XMSWATCH:";

static CHAR header [] = "\n\
\n\
  Linear Address  Size (KB)  Locks   Handle	Owner	    Properties \n\
  ---------------------------------------------------------------------";

static void PutLocks (struct HANDLE *);
static void PutHandle (struct HANDLE *);
static void PutOwner (struct HANDLE *);
static void PutProperties (struct HANDLE *);

static void SpaceOut (void);

/*  A few functions developed for FREE.C help harmonise this report with the
    one about free memory blocks.  */

extern void PutLinearAddress (struct HANDLE *);
extern void PutSizeInKB (struct HANDLE *);
extern void PutShortfall (WORD);

extern WORD BuildListOfFreeEMBs (struct LIST *);

/*  Material needed for comparison of this report with the one that FREE.C
    generates  */

extern struct LIST *pFreeEMBs;

static int TestRegion (DWORD, WORD);

#define NONE_FREE   0
#define SOME_FREE   1
#define ALL_FREE    2

static BOOL IsAvailableHandle (WORD);

/*  ========================================================================  */

void ReportWatchedEMBs (void)
{
    register struct HANDLE *ptr;
    register WORD handle;
    WORD unaccounted;
    void (_far *xmsptr_save) (void);
    struct EMBINFO info;

    /*	Information about EMBs that have been allocated already can be
	obtained only by intercepting the XMS calls at the time of
	allocation to record the corresponding handle.	If the resident
	component of the program (the watcher) has been installed, retrieve
	its data.  */

    pWatchedEMBs = ConsultWatcher (&xmswatch);
    if (pWatchedEMBs == NULL) return;

    /*	One essential piece of data is the XMS calling address understood by
	the earlier XMSWATCH installation.  This is returned as one of the
	fields in the XMSWATCH structure and must replace lpXMS (the present
	XMS calling point) temporarily, so that questions about monitored
	handles can be directed to the XMS server that actually allocated
	the handles.

	Observe that the lpNextXMS field is the destination of the far jump
	that started the XMS handler hooked by the watcher.  The following
	code therefore undercuts subsequent additions to the XMS chain - it
	is technically illegal and it is essential that operations conducted
	this way be non-invasive.  Really, the onus is on the XMS monitoring
	code in MONITOR.C to watch more functions than the ones implemented
	for this demonstration version of the program.	*/

    xmsptr_save = lpXMS;
    lpXMS = xmswatch.lpNextXMS;

    /*	Each entry in the resident handle table (a copy of which is now at
	the address pointed to by pWatchedEMBs) contains just a handle and a
	string describing the DOS program that obtained the allocation.  For
	each handle monitored this way, obtain the base address, size and
	lock count of the EMB it denotes, then sort the entry in order of
	increasing base address.  */

    for (ptr = pWatchedEMBs -> pUnused; ptr < pWatchedEMBs -> pEnd; ptr ++) {
	if (handle = ptr -> wHandle) {
	    if (GetEMBInfo (handle, &info)) {
		ptr -> cLocks = info.cLocks;
		ptr -> wSizeInKB = info.wSizeInKB;
		ptr -> wFlags = USED_EMB;
		if (LockEMB (handle, &(ptr -> dwBase))) {
		    ptr -> wFlags |= PHYSICAL_EMB;
		    UnlockEMB (handle);
		}
		LinkEntry (pWatchedEMBs, ptr);
	    }
	}
    }

    /*	Add information about the free memory blocks as understood by the
	XMS server whose activity has been monitored, then restore lpXMS to
	the current XMS calling point.	*/

    unaccounted = BuildListOfFreeEMBs (pWatchedEMBs);

    lpXMS = xmsptr_save;

    /*	Display the results only if at least one non-trivial EMB (whether
	allocated or still free) has been discovered.  In the case where no
	block can be found (most likely because all extended memory was
	allocated before the watcher's installation), the list may be useful
	to a future extension of the program because it provides numerical
	values for the XMS server's spare handles.  */

    ptr = pWatchedEMBs -> pLowest;
    for (;;) {
	if (ptr == NULL) return;
	if (ptr -> wFlags & (USED_EMB | FREE_EMB)) break;
	ptr = ptr -> pNext;
    }

    printf (banner);
    printf ("\n\n  Handles monitored:  %u", xmswatch.wHandles);
    printf (header);

    /*	Describe the blocks in order of ascending linear address.  */

    for (ptr = pWatchedEMBs -> pLowest; ptr != NULL; ptr = ptr -> pNext) {
	if (ptr -> wFlags & (USED_EMB | FREE_EMB)) {
	    printf ("\n     ");
	    PutLinearAddress (ptr);
	    SpaceOut ();
	    PutSizeInKB (ptr);
	    SpaceOut ();
	    PutLocks (ptr);
	    SpaceOut ();
	    PutHandle (ptr);
	    SpaceOut ();
	    PutOwner (ptr);
	    SpaceOut ();
	    PutProperties (ptr);
	}
    }
    if (unaccounted) PutShortfall (unaccounted);
    printf ("\n");
}

/*  The Locks, Handle and Owner headings are relevant only for allocated
    EMBs.  For free blocks, maintain column alignment by padding the display
    with an equivalent number of spaces.  */

static void PutLocks (struct HANDLE *ptr)
{
    if (ptr -> wFlags & USED_EMB) printf ("%3u", ptr -> cLocks);
    else printf ("   ");
}

static void PutHandle (struct HANDLE *ptr)
{
    if (ptr -> wFlags & USED_EMB) printf ("%04Xh", ptr -> wHandle);
    else printf ("     ");
}

static void PutOwner (struct HANDLE *ptr)
{
    if (ptr -> wFlags & USED_EMB) printf ("%-8s", ptr -> cOwner);
    else printf ("        ");
}

static void SpaceOut (void)
{
    printf ("     ");
}

/*  The most straightforward purpose of the Properties heading is to avoid
    interfering with other columns when indicating that a block is free.
    This is only a question of taste however:  indeed, free blocks are
    distinguished by having no handle or owner (some allocated blocks may
    appear to have no owner because the XMS caller lay outside the DOS
    memory arena system at the time of allocation).

    A greater purpose is to compare the extended memory picture seen by the
    XMS server whose activity has been monitored by the resident XMSWATCH
    copy with the picture currently presented to applications.	Extensions
    to the XMS chain can produce very curious effects:	it is the job of the
    PutProperties function to report possible discrepancies.  */

static void PutProperties (register struct HANDLE *ptr)
{
    DWORD linear_address;
    struct EMBINFO info;

    /*	If the block described by ptr is free, say so.	*/

    if (ptr -> wFlags & FREE_EMB) {
	printf ("free");

	/*  Note however, that although the XMS server monitored by the
	    resident XMSWATCH copy regards the block as free, a subsequent
	    extension to the XMS chain might not be making the block
	    available.	*/

	if (ptr -> wFlags & PHYSICAL_EMB) {
	    if (TestRegion (ptr -> dwBase, ptr -> wSizeInKB) != ALL_FREE)
		printf (" but hidden");
	}
	return;
    }

    /*	A block of memory that has been allocated (but not yet released) at
	the level monitored by the resident copy of XMSWATCH might seem free
	to a program making XMS calls now.  A likely cause is that a program
	is redistributing memory to its clients, having hooked the XMS chain
	to control its clients' use of extended memory.  The condition can
	be detected only if some piece of the original allocation remains
	undistributed.	*/

    if (ptr -> wFlags & PHYSICAL_EMB) {
	if (ptr -> wSizeInKB != 0) {
	    if (TestRegion (ptr -> dwBase, ptr -> wSizeInKB) != NONE_FREE) {
		printf ("sub-let");
		return;
	    }
	}
    }

    /*	An XMS extension might also be redistributing handles obtained from
	the original XMS server.  This will be revealed if a handle already
	assigned to an EMB (and detected by the resident XMSWATCH copy) also
	turns up in the pFreeEMBs list (as one of the current XMS server's
	spares).  */

    if (IsAvailableHandle (ptr -> wHandle)) {
	printf ("consumed");
	return;
    }

    /*	Other discrepancies may be uncovered by asking legally for
	information about each handle - by issuing XMS calls to the whole
	XMS chain rather than undercutting to the handler monitored by the
	resident copy of XMSWATCH.  Extensions to the chain might not
	recognise a handle allocated before their addition or perhaps they
	associate the same handle with a different block of memory.

	Be aware though, that an XMS handle is normally a private matter
	between the XMS server and the client that obtains the allocation.
	XMSWATCH interferes with this relationship by eavesdropping on XMS
	activity.  A program that hooks the XMS chain after obtaining a
	handle to an EMB might react unfavourably to seeing that handle
	appear in an XMS request issued by another program.  This is the
	case with the enhanced mode of Windows 3.0, which deems this
	activity to violate system integrity - the consequences are surely
	harmless, but if the compatibility is important, the following
	conditional compilation block should be skipped.  */

    #if TRUE

    if (NOT GetEMBInfo (ptr -> wHandle, &info)) {
	printf ("invalidated");
	return;
    }
    if (ptr -> wFlags & PHYSICAL_EMB) {
	if (NOT LockEMB (ptr -> wHandle, &linear_address)) {
	    printf ("virtualised");
	    return;
	}
	UnlockEMB (ptr -> wHandle);
	if (ptr -> dwBase != linear_address
		OR ptr -> wSizeInKB != info.wSizeInKB) {
	    printf ("converted");
	    return;
	}
    }
    #endif
}

/*  Given a region of extended memory, the next function assesses whether
    all, any or none of the region is currently available through the XMS.
    The function relies on the pFreeEMBs list built by code in the FREE.C
    module.  */

static int TestRegion (DWORD test_base, WORD test_sizeinkb)
{
    register struct HANDLE *ptr;
    DWORD ptr_top, test_top = test_base + ((DWORD) test_sizeinkb << 10);
    int result = NONE_FREE;

    /*	Search the list that describes the blocks and handles seen at
	application level.  The list is arranged in increasing order of base
	address, so abort the search on finding the first block that lies
	entirely above the test region.  */

    ptr = pFreeEMBs -> pLowest;
    while (ptr != NULL AND ptr -> dwBase < test_top) {

	/*  Entries corresponding to spare handles are of no interest here
	    and neither are blocks of virtual memory for which a base
	    address could not be determined.  Some overlap between the
	    pFreeEMBs entry under consideration and the test region is also
	    required.  */

	if (ptr -> wFlags & FREE_EMB AND ptr -> wFlags & PHYSICAL_EMB) {
	    ptr_top = ptr -> dwBase + ((DWORD) ptr -> wSizeInKB << 10);
	    if (ptr_top > test_base) {

		/*  The simple case arises when the block described by ptr
		    contains the test region.  */

		if ( ptr -> dwBase <= test_base AND ptr_top >= test_top)
		    return (ALL_FREE);

		/*  Otherwise, the possibility remains that the test region
		    is the union of two or more blocks in the pFreeEMBs
		    list, so continue the algorithm for that part of the
		    test region that has not yet been covered.	*/

		test_base = ptr_top;
		result = SOME_FREE;
	    }
	}
	ptr = ptr -> pNext;
    }
    return (result);
}

static BOOL IsAvailableHandle (WORD handle)
{
    register struct HANDLE *ptr;

    for (ptr = pFreeEMBs -> pLowest; ptr != NULL; ptr = ptr -> pNext) {
	if (ptr -> wHandle == handle) return (TRUE);
    }
    return (FALSE);
}

/*  ************************************************************************  */

