
;   ************************************************************************
;   *				    xms.asm				   *
;   ************************************************************************

;   This file provides a C language interface for intercepting XMS calls.

.NOLIST
  INCLUDE		intpack.inc
  INCLUDE		keepsegs.inc
.LIST

.MODEL			SMALL, C, FARSTACK

;   It is assumed that the presence of a machine supporting extended memory
;   will have been established before enabling this module and consequently,
;   that the PUSHA and POPA instructions are available.

.286

;   =======================================================================

.FARDATA		KEEP

;   XMS handlers are installed by threading themselves into a chain without
;   disturbing the original entry point.  This ensures that the most recent
;   addition has first option on all XMS requests, even those issued by
;   programs activated earlier.

;   The address at which to call the next handler in the chain must be
;   provided externally.

			EXTERN	lpNextXMS:DWORD

;   Jump table for XMS filters:

			PUBLIC	XMSFilterTable
XMSFilterTable		LABEL	WORD

;   All procedures in the table have C language prototypes of the form

;	unsigned XMSFilter (struct FARPACK STACKSEG *);

;   Functions not provided in other modules default to one which is
;   transparent.

NEXTNUM 		=	00h
@XMSFilter		MACRO	num:REQ, command:REQ
  IF			num LT NEXTNUM
    .ERR		<@XMSFilter macro not in sequence>
  ENDIF
  IF			num GT NEXTNUM
    REPEAT		num - NEXTNUM
			dw	OFFSET KEEP_TEXT:XMSFilterPass
    ENDM
  ENDIF
  .CODE 		KEEP_TEXT
			EXTERN	XMSFilter&command (XMSFilterPass):NEAR
  .FARDATA		KEEP
			dw	OFFSET KEEP_TEXT:XMSFilter&command
  NEXTNUM		=	num + 01h
ENDM

@XMSFilter		00h, GetVersion
@XMSFilter		01h, AllocateHMA
@XMSFilter		02h, FreeHMA
@XMSFilter		03h, GlobalEnableA20
@XMSFilter		04h, GlobalDisableA20
@XMSFilter		05h, LocalEnableA20
@XMSFilter		06h, LocalDisableA20
@XMSFilter		07h, QueryA20
@XMSFilter		08h, QueryFreeMemory
@XMSFilter		09h, AllocateEMB
@XMSFilter		0Ah, FreeEMB
@XMSFilter		0Bh, MoveData
@XMSFilter		0Ch, LockEMB
@XMSFilter		0Dh, UnlockEMB
@XMSFilter		0Eh, GetHandleInfo
@XMSFilter		0Fh, ResizeEMB
@XMSFilter		10h, AllocateUMB
@XMSFilter		11h, FreeUMB

;   A private variable to support stack restitution by the XMSFilterPass
;   function

pStack			dw	?

;   ========================================================================

.CODE			KEEP_TEXT

xms			PROC	FAR PUBLIC

;   XMS handlers begin with the following instruction sequence that may be
;   replaced by an intersegment jump should another handler need to be
;   installed.

			jmp	short	@f

			nop
			nop
			nop

@@:

;   Save the caller's registers on the stack.  As with _interrupt functions
;   (for which the same structure is used), this ensures the preservation of
;   registers across the XMS call unless deliberate action.  To support the
;   return of different values in the registers, the XMSFilter functions are
;   all passed a stack-based pointer to the register copy.

			pusha
			push	ds
			push	es

			mov	bp,sp

;   Although the CPU flags are not included in the structure of stacked
;   registers, they must also be saved across the XMS call.

			pushf

;   Load ds with the segment address of the resident data group.

			mov	dx,KEEP_GROUP
			mov	ds,dx
			ASSUME	ds:KEEP_GROUP

;   Save the stack pointer to allow the XMSFilterPass function to recover
;   the entry conditions.

			push	[pStack]		; for re-entrancy
			mov	[pStack],sp

;   Safeguard against incomplete initialisation.  This supports more
;   generality in the routine that patches in the far jump when hooking the
;   XMS:  without this provision, that routine would need to have knowledge
;   of the lpNextXMS variable's address to initialise it before the patch.
;   Thus, in the short time between hooking into the XMS chain and setting
;   lpNextXMS (to reconnect the chain), return the "unsupported function"
;   error for all XMS calls in preference to having XMS requests go astray.

			mov	dx,word ptr [lpNextXMS]
			or	dx,word ptr [lpNextXMS + 02h]
			jnz	@f

			mov	(REGPACK PTR [bp]).stacked_ax,0000h
			mov	(REGPACK PTR [bp]).stacked_bl,80h

			jmp	done

@@:

;   Determine which procedure should handle the XMS function.

			mov	al,ah
			cmp	al,NEXTNUM
			jnb	too_high

			xor	ah,ah
			mov	si,ax
			shl	si,1
			mov	si,[si + XMSFilterTable]

			jmp	@f

too_high:
			mov	si,OFFSET XMSFilterPass

@@:

;   Pass each filter a pointer to the caller's registers saved on the stack.
;   The register values passed directly to the filter are all undefined,
;   except that ds = KEEP_GROUP.

			push	bp
			call	si
			pop	bx

done:

;   By returning normally, a filter indicates that it has completed the XMS
;   request.  Register values for the XMS caller will have been entered into
;   the image on the stack, so retrieve them and return.

			pop	[pStack]
			popf

			pop	es
			pop	ds
			popa

			ret

xms			ENDP

;   ------------------------------------------------------------------------

XMSFilterPass		PROC	PUBLIC

;   At any time, a filter may pass the XMS call to the next handler by
;   calling this function - much the same as using the _chain_intr function
;   in the Microsoft C library, except that in this case, the destination
;   for the control transfer is fixed and need not be passed as an argument.

;	void XMSFilterPass (void);

			mov	sp,[pStack]
			pop	[pStack]
			popf

			pop	es
			pop	ds
			popa

;   The address of the next handler is kept in KEEP_GROUP, which is
;   convenient for the C language interface, but makes an indirect jump
;   impossible without corrupting the XMS caller's segment registers.  Place
;   the address on the stack in such a way that a RETF transfers control
;   correctly.

			push	ax	; subtract four bytes from stack
			push	ax	; without disturbing the flags

			push	bp
			mov	bp,sp
			push	ds
			push	ax
			mov	ax,KEEP_GROUP
			mov	ds,ax
			lds	ax,[lpNextXMS]
			mov	[bp + 02h],ax	; place the transfer address
			mov	[bp + 04h],ds	; over the four extra bytes
			pop	ax
			pop	ds
			pop	bp

			retf

XMSFilterPass		ENDP

;   ------------------------------------------------------------------------

;   A function is provided for calling the next XMS handler in the chain.
;   Registers are taken from a structure on the stack (which need not be the
;   one in which the caller's registers were saved).

;	unsigned CallNextXMS (struct REGPACK STACKSEG *);

CallNextXMS		PROC	PUBLIC USES ds di si, regptr:WORD
			LOCAL	handler:DWORD

			les	bx,[lpNextXMS]

			mov	word ptr [handler],bx
			mov	word ptr [handler + 02h],es

;   Load registers with values for the XMS server.  In version 2.0 of the
;   specification, only registers ax, bx, dx, si and ds are ever meaningful.

			push	bp
			mov	bp,[regptr]
			ASSUME	bp:PTR INTPACK

                        mov     ds,[bp].stacked_ds
                        mov     si,[bp].stacked_si
			mov	dx,[bp].stacked_dx
			mov	bx,[bp].stacked_bx
			mov	ax,[bp].stacked_ax

			pop	bp
			ASSUME	bp:NOTHING

;   Call the required XMS handler.

			call	[handler]

;   Copy the return values onto the stack.  XMS version 2.0 returns values
;   in registers ax, bx and dx only.  The implementation in HIMEM.SYS
;   preserves cx, si, di, ds and es explicitly - it is also capable of
;   corrupting values in bx and dx even if the specification assigns them
;   no meaning.

			push	bp
			mov	bp,[regptr]
			ASSUME	bp:PTR INTPACK

			mov	[bp].stacked_ax,ax
			mov	[bp].stacked_bx,bx
			mov	[bp].stacked_dx,dx

			pop	bp
			ASSUME	bp:NOTHING

			ret

CallNextXMS		ENDP

;   ************************************************************************

END

